from ninja import NinjaAPI, Schema

from allianceauth.services.hooks import get_extension_logger

from incursions.models.waitlist import (
    ApprovedFitting, WaitlistCategory, WaitlistCategoryRule,
)


class DNAFittingSchema(Schema):
    name: str
    dna: str
    tier: str
    implant_set: str


class FittingNoteSchema(Schema):
    name: str
    description: str


class FittingResponse(Schema):
    fittingdata: list[DNAFittingSchema] | None
    notes: list[FittingNoteSchema] | None
    logi_rules: list[int] | None  # Im hoping to deprecate this entirely, its dumb


logger = get_extension_logger(__name__)

api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    FittingsAPIEndpoints(api)


class FittingsAPIEndpoints:

    tags = ["Fittings"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/fittings", response={200: FittingResponse, 403: dict, 404: dict}, tags=self.tags)
        def fittings(request):
            if not request.user.has_perm("incursions.basic_waitlist"):
                return 403, {"error": "Permission denied"}

            fittings = ApprovedFitting.objects.all()

            logi_category, _ = WaitlistCategory.objects.get_or_create(name="LOGI")  # Fuck you, have a logi category even if you dont want it

            return FittingResponse(
                fittingdata=fittings,
                notes=fittings,
                logi_rules=WaitlistCategoryRule.objects.filter(waitlist_category=logi_category).values_list("ship__type_id", flat=True)
            )



from datetime import datetime, timezone

from ninja import NinjaAPI, Schema

from django.shortcuts import get_object_or_404

from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_main_character_from_user
from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import CharacterSchema
from incursions.models.waitlist import CharacterRoles, Role


class RoleSchema(Schema):
    name: str
    description: str


class CommanderSchema(Schema):
    character: CharacterSchema
    role: RoleSchema
    granted_by: CharacterSchema | None
    granted_at: datetime


class CommanderFiltersSchema(Schema):
    name: str
    member_count: int


class CommanderListSchema(Schema):
    commanders: list[CommanderSchema]
    filters: list[CommanderFiltersSchema]


class RequestPayload(Schema):
    character_id: int
    role: str


logger = get_extension_logger(__name__)

api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    CommandersAPIEndpoints(api)

# Hey Ariel, A commander is just a Character with Roles ++


class CommandersAPIEndpoints:

    tags = ["Commanders"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/commanders", response={200: CommanderListSchema, 403: dict}, tags=self.tags)
        def list_commanders(request):
            if not request.user.has_perm("incursions.view_commanders"):
                return 403, {"error": "Permission denied"}

            return 200, CommanderListSchema(commanders=CharacterRoles.objects.all(), filters=Role.objects.all())

        @api.post("/commanders", response={200: dict, 400:dict, 403: dict, 404: dict}, tags=self.tags)
        def assign_commander(request, payload: RequestPayload):
            if not request.user.has_perm("incursions.manage_commanders"):
                return 403, {"error": "Permission denied"}

            character = get_object_or_404(EveCharacter, character_id=payload.character_id)
            if CharacterRoles.objects.filter(character=character).exists():
                return 400, {"error": "Character already has a role"}

            role = get_object_or_404(Role, name=payload.role)
            if role.power > CharacterRoles.objects.get(character=get_main_character_from_user(request.user)).role.power:
                return 403, {"error": "Permission denied, You dont have the power to assign this role"}

            commander = CharacterRoles.objects.create(
                character=character,
                role=role,
                granted_by=get_main_character_from_user(request.user),
                granted_at=datetime.now(timezone.utc)
            )
            return 200, {"status": "Commander assigned", "id": commander.pk}

        @api.get("/commanders/roles", response={200: list[str], 403: dict}, tags=self.tags)
        def assignable_roles(request):
            if not request.user.has_perm("incursions.view_commanders"):
                return 403, {"error": "Permission denied"}

            return 200, list(Role.objects.all().values_list("name", flat=True))

        @api.get("/commanders/{character_id}", response={200: str, 403: dict}, tags=self.tags)
        def lookup_commander(request, character_id: int):
            if not request.user.has_perm("incursions.view_commanders"):
                return 403, {"error": "Permission denied"}

            character = get_object_or_404(EveCharacter, character_id=character_id)
            character_role = get_object_or_404(CharacterRoles, character=character)

            return 200, character_role.role.name

        @api.delete("/commanders/{character_id}", response={200: dict, 403: dict, 404: dict}, tags=self.tags)
        def revoke_commander(request, character_id: int):
            if not request.user.has_perm("incursions.manage_commanders"):
                return 403, {"error": "Permission denied"}

            character = get_object_or_404(EveCharacter, character_id=character_id)
            commander = get_object_or_404(CharacterRoles, character=character)

            if commander.role.power > CharacterRoles.objects.get(character=get_main_character_from_user(request.user)).role.power:
                return 403, {"error": "Permission denied, You dont have the power to rovoke this role"}

            try:
                deleted, _ = CharacterRoles.objects.filter(character_id=character_id).delete()
            except CharacterRoles.DoesNotExist:
                return 404, {"error": "Commander role not found"}

            return 200, {"status": "Commander role revoked"}



import bz2
import json
import lzma
import os

from eveuniverse.models import EveConstellation, EveSolarSystem

from django.db import IntegrityError
from django.utils import timezone

from allianceauth.eveonline.models import EveFactionInfo
from allianceauth.services.hooks import get_extension_logger

from incursions.models.incursion import Incursion, IncursionInfluence

from .static_data import incursion_constellations

logger = get_extension_logger(__name__)


EVEREF_STAGING_FOLDER = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'data.everef.net/incursions/history/')


def import_staging_history(dir: str = EVEREF_STAGING_FOLDER):
    for entry in sorted(os.scandir(dir), key=lambda e: e.name):
        if entry.name.startswith('incursions-') and entry.is_file():
            with bz2.open(entry) as f:
                data = f.read()
                magic_time = timezone.datetime.strptime(entry.name.removesuffix(".json.bz2").removeprefix(
                    "incursions-"), '%Y-%m-%d_%H-%M-%S').replace(tzinfo=timezone.utc)
                print(magic_time)
                # 2024-04-06_00-00-01
                # %Y-%m-%d_%H-%M-%S
                for incursion in json.loads(data):
                    # [{"constellation_id":20000651,"faction_id":500019,"has_boss":false,"infested_solar_systems":[30004459,30004460,30004461,30004462,30004463,30004464],"influence":0.0,"staging_solar_system_id":30004464,"state":"mobilizing","type":"Incursion"},{"constellation_id":20000520,"faction_id":500019,"has_boss":true,"infested_solar_systems":[30003565,30003566,30003567,30003568,30003569,30003570,30003571,30003572],"influence":1.0,"staging_solar_system_id":30003572,"state":"mobilizing","type":"Incursion"},{"constellation_id":20000173,"faction_id":500019,"has_boss":false,"infested_solar_systems":[30001184,30001185,30001186,30001187,30001182,30001183],"influence":0.0,"staging_solar_system_id":30001186,"state":"mobilizing","type":"Incursion"},{"constellation_id":20000618,"faction_id":500019,"has_boss":false,"infested_solar_systems":[30004224,30004225,30004226,30004227,30004228,30004229,30004222,30004223],"influence":0.32516667656600473,"staging_solar_system_id":30004227,"state":"established","type":"Incursion"}]
                    # I have to apply my own logic to essentially _backdate_ the inserts at a historical point in time.
                    if incursion['state'] == "established":
                        try:
                            i = Incursion.objects.get(
                                established_timestamp__date__gte=magic_time.date() - timezone.timedelta(days=5),
                                established_timestamp__date__lte=magic_time.date() + timezone.timedelta(days=5),
                                constellation=EveConstellation.objects.get_or_create_esi(
                                    id=incursion['constellation_id'])[0])
                            continue
                        except Incursion.DoesNotExist:
                            i = Incursion.objects.create(
                                constellation=EveConstellation.objects.get_or_create_esi(
                                    id=incursion['constellation_id'])[0],
                                faction=EveFactionInfo.objects.get_or_create(faction_id=incursion['faction_id'])[0],
                                has_boss=True if incursion['has_boss'] == "true" else False,
                                staging_solar_system=EveSolarSystem.objects.get_or_create_esi(
                                    id=incursion['staging_solar_system_id'])[0],
                                state=incursion['state'],
                                type=incursion['type'],
                                established_timestamp=magic_time
                            )
                        except Incursion.MultipleObjectsReturned as e:
                            # I dont know how to handle this, im hoping if i ignore it, the other data makes up for it.
                            logger.exception(e)
                            continue

                    elif incursion['state'] == "mobilizing":
                        try:
                            i = Incursion.objects.get(
                                mobilizing_timestamp__date__gte=magic_time.date() - timezone.timedelta(days=2),
                                mobilizing_timestamp__date__lte=magic_time.date() + timezone.timedelta(days=2),
                                constellation=EveConstellation.objects.get_or_create_esi(
                                    id=incursion['constellation_id'])[0])
                            continue
                        except Incursion.DoesNotExist:
                            try:
                                i = Incursion.objects.get(
                                    established_timestamp__date__gte=magic_time.date() - timezone.timedelta(days=7),
                                    established_timestamp__date__lte=magic_time.date(),
                                    constellation=EveConstellation.objects.get_or_create_esi(
                                        id=incursion['constellation_id'])[0])
                                i.mobilizing_timestamp = magic_time
                                if i.state is not Incursion.States.ENDED:
                                    i.state = incursion['state']
                                if i.has_boss == "true":
                                    i.has_boss = True
                                i.save()
                            except Incursion.DoesNotExist:
                                print(f"Incursion Not able to be set Mobilizing {magic_time} {incursion}")
                        except Incursion.MultipleObjectsReturned as e:
                            # I dont know how to handle this, im hoping if i ignore it, the other data makes up for it.
                            logger.exception(e)
                            continue
                    elif incursion['state'] == "withdrawing":
                        try:
                            i = Incursion.objects.get(
                                withdrawing_timestamp__date__gte=magic_time.date() - timezone.timedelta(days=1),
                                withdrawing_timestamp__date__lte=magic_time.date(),
                                constellation=EveConstellation.objects.get_or_create_esi(
                                    id=incursion['constellation_id'])[0])
                            continue
                        except Incursion.DoesNotExist:
                            try:
                                i = Incursion.objects.get(
                                    established_timestamp__date__gte=magic_time.date() - timezone.timedelta(days=8),
                                    established_timestamp__date__lte=magic_time.date(),
                                    constellation=EveConstellation.objects.get_or_create_esi(
                                        id=incursion['constellation_id'])[0])
                                i.withdrawing_timestamp = magic_time
                                if i.state is not Incursion.States.ENDED:
                                    i.state = incursion['state']
                                if i.has_boss == "true":
                                    i.has_boss = True
                                i.save()
                            except Incursion.DoesNotExist:
                                try:
                                    i = Incursion.objects.get(
                                        mobilizing_timestamp__date__gte=magic_time.date() - timezone.timedelta(days=2),
                                        mobilizing_timestamp__date__lte=magic_time.date(),
                                        constellation=EveConstellation.objects.get_or_create_esi(
                                            id=incursion['constellation_id'])[0])
                                    i.withdrawing_timestamp = magic_time
                                    if i.state is not Incursion.States.ENDED:
                                        i.state = incursion['state']
                                    if i.has_boss == "true":
                                        i.has_boss = True
                                    i.save()
                                except Incursion.DoesNotExist:
                                    print(f"Incursion Not able to be set Withdrawing {magic_time} {incursion}")
                        except Incursion.MultipleObjectsReturned as e:
                            # I dont know how to handle this, im hoping if i ignore it, the other data makes up for it.
                            logger.exception(e)
                            continue
                try:
                    endeds = Incursion.objects.exclude(
                        state=Incursion.States.ENDED).filter(
                        established_timestamp__lte=magic_time)
                    for ended in endeds:
                        ended.state = Incursion.States.ENDED
                        ended.ended_timestamp = magic_time
                        ended.save()
                except Exception:
                    pass

        elif entry.is_dir():
            import_staging_history(entry.path)


def import_staging_backfill(dir: str = EVEREF_STAGING_FOLDER):
    with lzma.open(os.path.join(EVEREF_STAGING_FOLDER, "backfills/eve-incursions-de-2023-10-12.json.xz")) as f:
        data = f.read()
        for incursion in json.loads(data):
            eve_constellation = EveConstellation.objects.get_or_create_esi(
                id=incursion['spawn']["constellation"]["id"])[0]
            if incursion['spawn']['endedAt'] is None:
                print(f"Incursion not gracefully ended {incursion}")
                pass
            try:
                i = Incursion.objects.get(
                    established_timestamp__date=timezone.datetime.strptime(
                        str(incursion['spawn']["establishedAt"]), "%Y-%m-%dT%H:%M:%S.%fZ").date(),
                    ended_timestamp__date=timezone.datetime.strptime(
                        str(incursion['spawn']["endedAt"]), "%Y-%m-%dT%H:%M:%S.%fZ").date(),
                    constellation=eve_constellation)
            except Incursion.DoesNotExist:
                try:
                    i = Incursion.objects.create(
                        constellation=eve_constellation,
                        faction=EveFactionInfo.objects.get_or_create(faction_id=500019)[0],
                        staging_solar_system=EveSolarSystem.objects.get(
                            name=incursion_constellations[eve_constellation.name]["Staging"]),
                        state=Incursion.States.ENDED,
                        established_timestamp=timezone.datetime.strptime(
                            str(incursion['spawn']["establishedAt"]), "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=timezone.utc),
                        ended_timestamp=timezone.datetime.strptime(
                            str(incursion['spawn']["endedAt"]), "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=timezone.utc)
                    )
                except EveSolarSystem.DoesNotExist:
                    logger.error(eve_constellation.name)
            # Save an extra timestamps we can gather from this data
            try:
                if incursion['state'] == "Mobilizing":
                    i.mobilizing_timestamp = timezone.datetime.strptime(
                        str(incursion['date']), "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=timezone.utc)
                    i.save()
                elif incursion['state'] == "Withdrawing":
                    i.withdrawing_timestamp = timezone.datetime.strptime(
                        str(incursion['date']), "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=timezone.utc)
                    i.save()
            except Exception as e:
                logger.exception(e)
            # Moving onto IncursionInfluence before we close the loop
            for ilog in incursion["spawn"]["influenceLogs"]:
                try:
                    IncursionInfluence.objects.create(
                        incursion=i,
                        influence=ilog['influence'],
                        timestamp=timezone.datetime.strptime(str(ilog["date"]), "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=timezone.utc))
                except IntegrityError:
                    # We already have this data
                    pass

#   {
#     "id": "1",
#     "date": "2015-02-06T18:13:45.000Z",
#     "state": "Established",
#     "spawn": {
#       "id": "12",
#       "establishedAt": "2015-02-06T18:13:45.000Z",
#       "endedAt": "2015-02-09T16:00:01.000Z",
#       "state": "Ended",
#       "constellation": {
#         "id": "20000011"
#       },
#       "influenceLogs": [
#         {
#           "id": "293702",
#           "date": "2021-10-30T09:00:00.000Z",
#           "influence": 0
#         },
#         {
#           "id": "293707",
#           "date": "2021-10-30T10:00:00.000Z",
#           "influence": 0
#         },
#         {
#           "id": "293712",
#           "date": "2021-10-30T11:00:00.000Z",
#           "influence": 0.048
#         },
#         {
#           "id": "293717",
#           "date": "2021-10-30T12:00:00.000Z",
#           "influence": 0.175333
#         },
#     }
#   },



import base64
import json
import struct

import msgpack
import requests
from nacl.bindings import crypto_aead_xchacha20poly1305_ietf_encrypt
from nacl.utils import random as nacl_random
from pydantic import BaseModel

from django.utils.timezone import now

from incursions import app_settings


class SSEClient:
    def __init__(self, url: str, secret: str) -> None:
        self.url = url
        self.key = bytes.fromhex(secret) if isinstance(secret, str) else secret
        self.VERSION_BYTE = b'\xBA'

    def branca_encode(self, payload: bytes, timestamp: int | None = None) -> str:
        if len(app_settings.SSE_SECRET) != 32:
            raise ValueError("Branca key must be exactly 32 bytes")

        nonce = nacl_random(24)
        ts_bytes = struct.pack(">I", int(timestamp or now().timestamp()))
        header = self.VERSION_BYTE + ts_bytes + nonce
        key = app_settings.SSE_SECRET.encode()

        ciphertext = crypto_aead_xchacha20poly1305_ietf_encrypt(payload, header, nonce, key)
        token = header + ciphertext
        return base64.urlsafe_b64encode(token).rstrip(b"=").decode()

    def events_url(self, topics: list[str]) -> str:
        request_payload = {"topics": topics}
        payload: bytes = msgpack.packb(request_payload, use_bin_type=True)
        token = self.branca_encode(payload, self.key)
        return f"{self.url}/events?token={token}"

    def submit(self, events: list["SSEEvent"]) -> requests.Response:
        submission = {"events": [event.to_dict() for event in events]}
        payload: bytes = msgpack.packb(submission, use_bin_type=True)
        encoded = self.branca_encode(payload)
        response = requests.post(f"{self.url}/submit", data=encoded)
        return response


class SSEEvent:
    def __init__(self, topic: str, event: str, data: str) -> None:
        self.topic = topic
        self.event = event
        self.data = data

    @classmethod
    def new(cls, topic: str, event: str, data: str) -> "SSEEvent":
        return cls(topic, event, data)

    @classmethod
    def new_json(cls, topic: str, event: str, data: BaseModel | dict) -> "SSEEvent":
        encoded = json.dumps(data)
        return cls(topic, event, encoded)

    def to_dict(self) -> dict:
        return {"topic": self.topic, "event": self.event, "data": self.data}



from ninja import NinjaAPI, Schema


class StatsResponse(Schema):
    fleet_seconds_by_hull_by_month: dict[str, dict[str, float]]
    xes_by_hull_by_month: dict[str, dict[str, float]]
    fleet_seconds_by_month: dict[str, float]
    pilots_by_month: dict[str, float]
    xes_by_hull_28d: dict[str, float]
    fleet_seconds_by_hull_28d: dict[str, float]
    x_vs_time_by_hull_28d: dict[str, dict[str, float]]
    time_spent_in_fleet_by_month: dict[str, dict[str, float]]


api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    StatisticsAPIEndpoints(api)


class StatisticsAPIEndpoints:

    tags = ["Statistics"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/stats", response={200: StatsResponse, 403: dict}, tags=self.tags)
        def statistics(request):
            if not request.user.has_perm("incursions.stats_view"):
                return 403, {"error": "Permission denied"}

            return StatsResponse()
            # return StatsResponse(
            #     fleet_seconds_by_hull_by_month=fleet_seconds_by_hull_by_month,
            #     xes_by_hull_by_month=xes_by_hull_by_month,
            #     fleet_seconds_by_month=fleet_seconds_by_month,
            #     pilots_by_month=pilots_by_month,
            #     xes_by_hull_28d=xes_by_hull_28d,
            #     fleet_seconds_by_hull_28d=fleet_seconds_by_hull_28d,
            #     x_vs_time_by_hull_28d=x_vs_time_by_hull_28d,
            #     time_spent_in_fleet_by_month=time_spent_in_fleet_by_month
            # )



from ninja import NinjaAPI, Schema

from allianceauth.framework.api.user import (
    get_all_characters_from_user, get_main_character_from_user,
)
from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import CharacterSchema


class WhoamiResponse(Schema):
    account_id: int
    access: list[str]
    characters: list[CharacterSchema]


logger = get_extension_logger(__name__)

api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    AuthAPIEndpoints(api)


class AuthAPIEndpoints:

    tags = ["Auth"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/auth/whoami", response={200: WhoamiResponse, 404: dict}, tags=self.tags)
        def whoami(request):
            user = request.user

            try:
                main_character = get_main_character_from_user(user)
            except Exception:
                return 404, {"error": "User has no Main Character Set"}

            access_levels: list[str] = []
            for perm in user.get_all_permissions():
                if perm.startswith("incursions."):
                    access_levels.append(perm)

            return 200, WhoamiResponse(account_id=main_character.character_id, access=access_levels, characters=[CharacterSchema.from_orm(alt) for alt in get_all_characters_from_user(user)])

        @api.get("/auth/logout", tags=self.tags)
        def logout(request) -> dict:
            # DEPRECATE THIS
            return {"status": "Logged out"}

        @api.get("/auth/login_url", tags=self.tags)
        def login_url(request, alt: bool = False, fc: bool = False) -> dict:
            state = "alt" if alt else "normal"
            scopes = ["publicData", "skills.read_skills", "clones.read_implants"]
            if fc:
                scopes.extend(["fleets.read_fleet", "fleets.write_fleet", "ui.open_window", "search"])

            return {
                "login_url": f"https://login.eveonline.com/v2/oauth/authorize?response_type=code&redirect_uri=YOUR_REDIRECT_URI&client_id=YOUR_CLIENT_ID&scope={' '.join(scopes)}&state={state}"
            }



from ninja import NinjaAPI, Schema

from django.shortcuts import get_object_or_404

from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_all_characters_from_user
from allianceauth.services.hooks import get_extension_logger

from incursions.providers import get_character_implants


class ImplantsResponse(Schema):
    implants: list[int]


logger = get_extension_logger(__name__)
api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    ImplantsAPIEndpoints(api)


class ImplantsAPIEndpoints:

    tags = ["Implants"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/implants", response={200: ImplantsResponse, 403: dict}, tags=self.tags)
        def list_implants(request, character_id: int):
            character = get_object_or_404(EveCharacter, character_id=character_id)
            if not (character in get_all_characters_from_user(request.user) or request.user.has_perm("incursions.implants_view")):
                return 403, {"error": "Permission denied"}

            implants, _ = get_character_implants(character_id)

            return ImplantsResponse(implants=implants)



from corptools.models import EveItemType
from solo.models import SingletonModel

from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from allianceauth.eveonline.models import (
    EveAllianceInfo, EveCharacter, EveCorporationInfo,
)

from incursions.providers import get_fleet_members, get_fleet_wings


class Role(models.Model):
    name = models.CharField(max_length=64, unique=True)
    description = models.TextField(null=True, blank=True)
    power = models.IntegerField(default=0, help_text="Bigger is better, user can assign a power level lower than this")

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Role")
        verbose_name_plural = _("Waitlist - Roles")

    def __str__(self) -> str:
        return f"{self.name}"

    @property
    def member_count(self) -> int:
        return CharacterRoles.objects.filter(role=self).count()


class Announcement(models.Model):

    message = models.TextField()
    is_alert = models.BooleanField(default=False)
    pages = models.TextField(null=True, blank=True)  # DEPRECATE THIS

    created_by = models.ForeignKey(
        EveCharacter,
        on_delete=models.SET_NULL, null=True, blank=True, related_name='+')
    created_at = models.DateTimeField(auto_now_add=True)
    revoked_by = models.ForeignKey(
        EveCharacter,
        on_delete=models.SET_NULL, null=True, blank=True, related_name='+')
    revoked_at = models.DateTimeField(default=None, blank=True, null=True)

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Announcement")
        verbose_name_plural = _("Waitlist - Announcements")

    def __str__(self) -> str:
        return f"Announcement #{self.pk}"


class Ban(models.Model):
    class entity_choices(models.TextChoices):
        CHARACTER = 'Character', _('Character')
        CORPORATION = 'Corporation', _('Corporation')
        ALLIANCE = 'Alliance', _('Alliance')

    entity_type = models.CharField(max_length=12, choices=entity_choices.choices, default=entity_choices.CHARACTER)
    entity_character = models.ForeignKey(
        EveCharacter,
        on_delete=models.CASCADE, null=True, blank=True, related_name='+')
    entity_corporation = models.ForeignKey(
        EveCorporationInfo,
        on_delete=models.CASCADE, null=True, blank=True, related_name='+')
    entity_alliance = models.ForeignKey(
        EveAllianceInfo,
        on_delete=models.CASCADE, null=True, blank=True, related_name='+')

    public_reason = models.CharField(max_length=512, null=True, blank=True)
    reason = models.CharField(max_length=512)

    issued_at = models.DateTimeField(auto_now_add=True)
    issued_by = models.ForeignKey(
        EveCharacter, on_delete=models.SET_NULL, null=True, blank=True, related_name='+')
    revoked_at = models.DateTimeField(blank=True, null=True, default=None)
    revoked_by = models.ForeignKey(
        EveCharacter, on_delete=models.SET_NULL,null=True,blank=True, related_name='+')

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Ban")
        verbose_name_plural = _("Waitlist - Bans")

    def __str__(self) -> str:
        return f"Ban #{self.pk} on {self.entity_type} {self.pk}"

    @property
    def entity_name(self) -> str:
        if self.entity_type == self.entity_choices.CHARACTER:
            return self.entity_character.character_name
        elif self.entity_type == self.entity_choices.CORPORATION:
            return self.entity_corporation.corporation_name
        elif self.entity_type == self.entity_choices.ALLIANCE:
            return self.entity_alliance.alliance_name
        return "Unknown Name"


class Badge(models.Model):
    name = models.CharField(max_length=64, unique=True)
    exclude_badge = models.ForeignKey(
        'self',
        on_delete=models.SET_NULL,null=True,blank=True,related_name='excluded_by')  # Mutually Exclusive Badges i think?

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Badge")
        verbose_name_plural = _("Waitlist - Badges")

    def __str__(self) -> str:
        return f"{self.name}"

    @property
    def member_count(self) -> int:
        return CharacterBadges.objects.filter(badge=self).count()


class CharacterRoles(models.Model):
    character = models.OneToOneField(
        EveCharacter,
        on_delete=models.CASCADE, primary_key=True)
    role = models.ForeignKey(
        Role,
        verbose_name=_("Incursion Role"), on_delete=models.CASCADE, related_name='incursion_role')

    granted_at = models.DateTimeField(auto_now_add=True)
    granted_by = models.ForeignKey(
        EveCharacter,
        on_delete=models.SET_NULL, blank=True, null=True, related_name='+')

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Character Role")
        verbose_name_plural = _("Waitlist - Character Roles")

    def __str__(self) -> str:
        return f"Role: {self.role} for {self.character}"


class CharacterBadges(models.Model):

    character = models.ForeignKey(
        EveCharacter,
        on_delete=models.CASCADE, related_name='+')
    badge = models.ForeignKey(Badge, on_delete=models.CASCADE, related_name='+')

    granted_at = models.DateTimeField(auto_now_add=True)
    granted_by = models.ForeignKey(
        EveCharacter,
        on_delete=models.SET_NULL, blank=True, null=True, related_name='+')

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Character Badge")
        verbose_name_plural = _("Waitlist - Character Badges")

    def __str__(self) -> str:
        return f"{self.character} has badge {self.badge}"


class CharacterNote(models.Model):

    character = models.ForeignKey(
        EveCharacter,
        on_delete=models.CASCADE,related_name='+')
    note = models.TextField()

    author = models.ForeignKey(
        EveCharacter,
        null=True, on_delete=models.SET_NULL, related_name='+')

    logged_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Character Note")
        verbose_name_plural = _("Waitlist - Character Note")

    def __str__(self) -> str:
        return f"Note #{self.pk} for {self.character}"


class ApprovedImplantSet(models.Model):
    class set_choices(models.TextChoices):  # ARIEL This is not Dynamic on the Frontend, hence CHOICES
        AMULET = "Amulet", _("Amulet")
        ASCENDENCY = "Ascendency", _("Ascendency")
        NONE = "None", _("None")  # Helpful for just DPS hardwirings or something idk

    name = models.CharField(max_length=255)
    set = models.CharField(max_length=16, choices=set_choices.choices, default=set_choices.AMULET)
    implants = models.CharField(max_length=255, unique=True)

    class Meta:
        verbose_name = _("Waitlist - Approved Implant Set")
        verbose_name_plural = _("Waitlist - Approved Implant Sets")

    def __str__(self) -> str:
        return self.implants


class ApprovedFitting(models.Model):
    class tier_choices(models.TextChoices):  # ARIEL This is not Dynamic on the Frontend, hence CHOICES
        BASIC = "Basic", _("Sponge")
        MAINLINE = "Mainline", _("Mainline")
        ALT = "Alt", _("Alt")
        OTHER = "Other", _("Other")

    ship = models.ForeignKey(
        EveItemType,
        verbose_name=_("Ship"), on_delete=models.CASCADE, related_name='+')
    name = models.CharField(max_length=255)
    dna = models.TextField(max_length=1024, unique=True)
    tier = models.CharField(max_length=16, choices=tier_choices.choices, default=tier_choices.MAINLINE)
    description = models.CharField(max_length=255, null=True, blank=True)
    implants = models.ForeignKey(
        ApprovedImplantSet,
        on_delete=models.SET_NULL, related_name='+', null=True, blank=True)

    class Meta:
        verbose_name = _("Waitlist - Approved Fitting")
        verbose_name_plural = _("Waitlist - Approved Fittings")

    def __str__(self) -> str:
        return self.name

    def is_logi(self) -> bool:
        return WaitlistCategoryRule.objects.get(ship=self.ship).waitlist_category.name == "LOGI"  # ARIEL this is a hardcoded requirement by frontend, not sure i like it

    def implant_set(self) -> str:
        return self.implants.set if self.implants else "None"


class Fitting(models.Model):
    ship = models.ForeignKey(
        EveItemType,
        verbose_name=_("Ship"), on_delete=models.CASCADE, related_name='+')
    dna = models.TextField(max_length=1024, unique=True)

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Waitlist Fit")
        verbose_name_plural = _("Waitlist - Waitlist Fits")

    def __str__(self) -> str:
        return f"Fitting {self.pk} (Hull: {self.ship})"


class ImplantSet(models.Model):
    implants = models.CharField(max_length=255, unique=True)

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Waitlist Implant Set")
        verbose_name_plural = _("Waitlist - Waitlist Implant Sets")

    def __str__(self) -> str:
        return f"ImplantSet #{self.pk}: {self.implants}"


class FittingHistory(models.Model):
    character = models.ForeignKey(EveCharacter, on_delete=models.CASCADE, related_name='+')
    fit = models.ForeignKey(Fitting, on_delete=models.CASCADE, related_name='+')
    implant_set = models.ForeignKey(ImplantSet, on_delete=models.CASCADE, related_name='+')

    logged_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Fitting History")
        verbose_name_plural = _("Waitlist - Fitting Histories")

    def __str__(self):
        return f"FitHistory #{self.pk} for {self.character}"


class SkillHistory(models.Model):
    # Faux this whole return, i really dont think i care about this
    character = models.ForeignKey(EveCharacter,on_delete=models.CASCADE, related_name='+')
    skill = models.ForeignKey(EveItemType, verbose_name=_("Skill"), on_delete=models.CASCADE, related_name='+')
    old_level = models.IntegerField(validators=[MinValueValidator(0), MaxValueValidator(5)])
    new_level = models.IntegerField(validators=[MinValueValidator(0), MaxValueValidator(5)])

    logged_at = models.DateTimeField(auto_now_add=True)

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Skill History")
        verbose_name_plural = _("Waitlist - Skill Histories")

    def __str__(self) -> str:
        return f"SkillHistory #{self.pk} for {self.character}"


class SkillCheck(models.Model):
    skill = models.ForeignKey(EveItemType, verbose_name=_("Skill"), on_delete=models.CASCADE, related_name='+')

    min = models.IntegerField(validators=[MinValueValidator(0), MaxValueValidator(5)], default=3)
    elite = models.IntegerField(validators=[MinValueValidator(0), MaxValueValidator(5)], default=4)
    gold = models.IntegerField(validators=[MinValueValidator(0), MaxValueValidator(5)], default=5)

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Skill Check")
        verbose_name_plural = _("Waitlist - Skill Check")

    def __str__(self) -> str:
        return f"Skill: {self.skill} - Min: {self.min}, Elite: {self.elite}, Gold: {self.gold}"


class ApprovedSkills(models.Model):
    hull = models.OneToOneField(EveItemType, verbose_name=_("Hull"), on_delete=models.CASCADE, related_name='+')
    skill_checks = models.ManyToManyField(SkillCheck, verbose_name=_("Skill Checks"), related_name='+')

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Approved Skills")
        verbose_name_plural = _("Waitlist - Approved Skills")

    def __str__(self) -> str:
        return f"Hull: {self.hull}"


class WaitlistCategory(models.Model):
    name = models.CharField(max_length=255, unique=True)

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Category")
        verbose_name_plural = _("Waitlist - Categories")

    def __str__(self) -> str:
        return f"{self.name}"


class WaitlistCategoryRule(models.Model):

    waitlist_category = models.ForeignKey(
        WaitlistCategory, verbose_name=_("Waitlist Category"), on_delete=models.CASCADE, related_name='+')
    ship = models.OneToOneField(
        EveItemType, verbose_name=_("Ship"), on_delete=models.CASCADE, related_name='+')

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Category Rule")
        verbose_name_plural = _("Waitlist - Category Rules")

    def __str__(self) -> str:
        return f"{self.waitlist_category} - {self.ship}"


class Fleet(models.Model):
    boss = models.ForeignKey(EveCharacter, on_delete=models.CASCADE, related_name='+')
    is_updating = models.BooleanField(default=False)

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Fleet")
        verbose_name_plural = _("Waitlist - Fleets")

    def __str__(self) -> str:
        return f"Fleet #{self.pk}"

    def get_fleet_members(self):
        # /api/members/{character_id} uses me
        # But if i pass it through here I can dump some info to DB while im at it.
        fleet_members = get_fleet_members(self.boss.character_id, self.pk)
        for fleet_member in fleet_members:
            # Save or update fleet members in the DB
            pass
        return fleet_members

    def get_fleet_wings(self):
        # /api/members/{character_id} uses me
        # But if i pass it through here I can dump some info to DB while im at it.
        fleet_wings = get_fleet_wings(self.boss.character_id, self.pk)
        for wing in fleet_wings:
            # Save or update fleet Wings in the DB
            pass
        return fleet_wings


class ActiveFleet(SingletonModel):
    fleet = models.ForeignKey(Fleet, on_delete=models.CASCADE, related_name='+')

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Fleet (Active)")
        verbose_name_plural = _("Waitlist - Fleet (Active)")

    def __str__(self) -> str:
        return f"Active Fleet {self.fleet}"


class FleetActivity(models.Model):
    character = models.ForeignKey(EveCharacter,on_delete=models.CASCADE, related_name='+')
    ship = models.ForeignKey(EveItemType, verbose_name=_("Ship"), on_delete=models.CASCADE, related_name='+')
    fleet = models.ForeignKey(
        Fleet, verbose_name=_(""), on_delete=models.CASCADE)

    first_seen = models.DateTimeField(auto_now_add=True)
    last_seen = models.DateTimeField(auto_now=True)
    has_left = models.BooleanField(default=False)
    is_boss = models.BooleanField(default=False)

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Fleet Activity")
        verbose_name_plural = _("Waitlist - Fleet Activities")

    def __str__(self) -> str:
        return f"FleetActivity #{self.pk} for {self.character}"


class FleetSquad(models.Model):
    """
    This is not explicitly a list of Squads in the Fleet
    This is a mapping of Squads to their configured Category a Character will be invited to
    """
    fleet = models.ForeignKey(Fleet, on_delete=models.CASCADE, related_name='+')
    category = models.ForeignKey(
        WaitlistCategory,
        verbose_name=_("Waitlist Category"), on_delete=models.CASCADE, related_name='+')
    wing_id = models.BigIntegerField()
    squad_id = models.BigIntegerField()

    class Meta:
        default_permissions = ()
        unique_together = (('fleet', 'category'),)
        verbose_name = _("Waitlist - Fleet Squad")
        verbose_name_plural = _("Waitlist - Fleet Squad")

    def __str__(self) -> str:
        return f"FleetSquad in Fleet {self.pk}, category {self.category}"

# The following Models are Specifically for the ACTIVE Waitlist, they are not historical
# They are used to manage the current state of the waitlist and its members


class Waitlist(SingletonModel):

    name = models.CharField(max_length=255, default="Waitlist")
    is_open = models.BooleanField(default=False)

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Waitlist")
        verbose_name_plural = _("Waitlist - Waitlist")

    def __str__(self) -> str:
        return f"{self.name}"


class WaitlistEntry(models.Model):
    waitlist = models.ForeignKey(Waitlist, on_delete=models.CASCADE, related_name='+')
    main_character = models.ForeignKey(EveCharacter,on_delete=models.CASCADE, related_name='+')

    joined_at = models.DateTimeField(auto_now_add=True, blank=True, null=True)

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Waitlist Entry")
        verbose_name_plural = _("Waitlist - Waitlist Entry")

    def __str__(self) -> str:
        return f"WaitlistEntry {self.main_character.character_name}"


class WaitlistEntryFit(models.Model):
    """
    WailistEntry - Fit
    An Entry can contain a series of fits for boxers
    """
    character = models.ForeignKey(EveCharacter,on_delete=models.CASCADE, related_name='+')
    waitlist_entry = models.ForeignKey(WaitlistEntry, on_delete=models.CASCADE, related_name='+')

    fit = models.ForeignKey(Fitting,on_delete=models.CASCADE, related_name='+')
    implant_set = models.ForeignKey(ImplantSet,on_delete=models.CASCADE, related_name='+')

    approved = models.BooleanField(default=False)
    tags = models.CharField(max_length=255,help_text="Comma Seperated List")
    category = models.ForeignKey(WaitlistCategory, verbose_name=_("Waitlist Category"), on_delete=models.CASCADE, related_name='+')
    fit_analysis = models.TextField(null=True, blank=True, help_text="[] apparently")
    review_comment = models.TextField(null=True, blank=True)
    is_alt = models.BooleanField(default=False)
    messagexup = models.TextField(null=True, blank=True)

    cached_time_in_fleet = models.BigIntegerField()  # Saved here to avoid repeated stats calls

    class Meta:
        default_permissions = ()
        verbose_name = _("Waitlist - Waitlist Entry Fit")
        verbose_name_plural = _("Waitlist - Waitlist Entry Fit")

    def __str__(self) -> str:
        return f"WaitlistEntry: {self.waitlist_entry.main_character.character_name} Fit: {self.character.character_name} - {self.fit.ship.name}"



from ninja import NinjaAPI, Schema

from allianceauth.services.hooks import get_extension_logger

from incursions.models.waitlist import WaitlistCategory


class CategorySchema(Schema):
    id: int
    name: str


api = NinjaAPI()


logger = get_extension_logger(__name__)


def setup(api: NinjaAPI) -> None:
    CategoriesAPIEndpoints(api)


class CategoriesAPIEndpoints:

    tags = ["Categories"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/categories", response={200: list[CategorySchema], 403: dict}, tags=self.tags)
        def list_categories(request):
            if not request.user.has_perm("incursions.basic_waitlist"):
                return 403, {"error": "Permission denied"}

            return 200, [CategorySchema.from_orm(cat) for cat in WaitlistCategory.objects.all()]



from allianceauth import __version__ as aa__version__
from allianceauth.services.hooks import get_extension_logger
from esi import __version__ as esi__version__
from esi.clients import EsiClientProvider
from esi.models import Token

from . import __version__

logger = get_extension_logger(__name__)

APP_INFO_TEXT = f"aa-incursions/{__version__} allianceauth/{aa__version__} django-esi/{esi__version__}"

"""
Swagger spec operations:
get_incursions
"""

esi = EsiClientProvider(app_info_text=APP_INFO_TEXT)


def get_incursions_incursions():
    operation = esi.client.Incursions.get_incursions()
    operation.request_config.also_return_response = True
    incursions, response = operation.results()
    return incursions, response


def get_fleet_wings(boss_character_id: int, fleet_id: int):
    required_scopes = ['esi-fleets.read_fleet.v1']
    token = Token.get_token(boss_character_id, required_scopes)

    result = esi.client.Fleets.get_fleets_fleet_id_wings(
        fleet_id=fleet_id,
        token=token.valid_access_token()
    ).results()
    return result


def get_fleet_members(boss_character_id: int, fleet_id: int):
    required_scopes = ['esi-fleets.read_fleet.v1']
    token = Token.get_token(boss_character_id, required_scopes)

    result = esi.client.Fleets.get_fleets_fleet_id_members(
        fleet_id=fleet_id,
        token=token.valid_access_token()
    ).results()
    return result


def kick_fleet_member(boss_character_id:int, fleet_id:int, character_id:int):
    required_scopes = ['esi-fleets.write_fleet.v1']

    token = Token.get_token(boss_character_id, required_scopes)

    result = esi.client.Fleets.delete_fleets_fleet_id_members_member_id(
        fleet_id=fleet_id,
        member_id=character_id,
        token=token.valid_access_token()
    ).results()
    return result


def kick_all_fleet_members(boss_character_id:int, fleet_id:int) -> int:
    '''
    Helper function to kick all fleet members
    get_fleet_members() -> kick_fleet_member()
    '''
    fleet_members = get_fleet_members(boss_character_id, fleet_id)

    count = fleet_members.count()
    for fleet_member in fleet_members:
        kick_fleet_member(boss_character_id, fleet_id, fleet_member['character_id'])

    return count


def search_esi(character_id:int, search_string:str, search_categories:str, strict:bool = False):
    required_scopes = ['esi-search.search_structures.v1']
    token = Token.get_token(character_id, required_scopes)

    operation = esi.client.Search.get_search(
        character_id=character_id,
        categories=search_categories,
        search=search_string,
        strict=strict,
        token=token.valid_access_token()
    )
    operation.request_config.also_return_response = True
    result, response = operation.results()
    return result, response


def open_window_information(character_id:int, target_id:int):
    required_scopes = ['esi-ui.open_window.v1']
    token = Token.get_token(character_id, required_scopes)

    operation = esi.client.UserInterface.post_ui_openwindow_information(
        character_id=character_id,
        target_id=target_id,
        token=token.valid_access_token(),
    )
    operation.request_config.also_return_response = True
    result, response = operation.results()
    return result, response


def get_character_implants(character_id:int):
    required_scopes = ['esi-clones.read_implants.v1']
    token = Token.get_token(character_id, required_scopes)

    operation = esi.client.Clones.get_characters_character_id_implants(
        character_id=character_id,
        token=token.valid_access_token()
    )
    operation.request_config.also_return_response = True
    result, response = operation.results()
    return result, response


def invite_to_fleet(boss_character_id:int, fleet_id:int, character_id:int, squad_id:int | None, wing_id:int | None, role: str = "squad_member"):
    required_scopes = ['esi-fleets.write_fleet.v1']
    token = Token.get_token(boss_character_id, required_scopes)

    operation = esi.client.Fleets.post_fleets_fleet_id_members(
        fleet_id=fleet_id,
        invitation={"character_id": character_id, "role": role},
        squad_id=squad_id if squad_id else 0,
        wing_id=wing_id if wing_id else 0,
        token=token.valid_access_token()
    )
    operation.request_config.also_return_response = True
    result, response = operation.results()
    return result, response



from datetime import datetime

from ninja import NinjaAPI, Schema

from django.utils.timezone import now

from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_main_character_from_user
from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import CharacterSchema
from incursions.models.waitlist import CharacterNote


class NoteSchema(Schema):
    character: CharacterSchema
    note: str
    author: CharacterSchema | None
    logged_at: datetime


class NotesSchema(Schema):
    notes: list[NoteSchema] | None


class AddNoteSchema(Schema):
    character_id: int
    note: str


logger = get_extension_logger(__name__)

api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    NotesAPIEndpoints(api)


class NotesAPIEndpoints:

    tags = ["Notes"]

    def __init__(self, api: NinjaAPI) -> None:

        # These endpoints were fucky, the client expects an emtpy list if no notes are found
        # and not a 404 error, so we return an empty list instead of a 404

        @api.get("/notes", response={200: NotesSchema, 403: dict}, tags=self.tags)
        def list_notes(request, character_id: int):
            if not request.user.has_perm("incursions.notes_view"):
                return 403, {"error": "Permission denied"}

            try:
                character = EveCharacter.objects.get(character_id=character_id)
                notes = CharacterNote.objects.filter(character=character)
            except EveCharacter.DoesNotExist:
                return NotesSchema(notes=[])
            except CharacterNote.DoesNotExist:
                return NotesSchema(notes=[])

            return NotesSchema(notes=notes)

        @api.post("/notes/add", tags=self.tags)
        def add_note(request, payload: AddNoteSchema):
            if not request.user.has_perm("incursions.notes_manage"):
                return 403, {"error": "Permission denied"}

            try:
                character = EveCharacter.objects.get(character_id=payload.character_id)
                note = CharacterNote(
                    character=character,
                    note=payload.note,
                    author=get_main_character_from_user(request.user),
                    logged_at=now()
                )
                note.save()
            except EveCharacter.DoesNotExist:
                return {"status": "Character not found"}
            except CharacterNote.DoesNotExist:
                return {"status": "Note not found"}

            return {"status": "OK"}



from django.db import models


class General(models.Model):
    """Meta model for app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("basic_waitlist", "Can Access The Incursion Waitlist Tool"),  # All
            ("waitlist_alts_view", "Can View all of a users Alts ! Danger?"),  # Leadership
            ("waitlist_announcements_manage", "can manage announcements"),  # Leadership
            ("waitlist_bages_view", "Can View Badges"),  # FC
            ("waitlist_badges_manage", "Can Manage Badges"),  # Leadership
            ("waitlist_bans_view", "Can View Bans"),  # FC
            ("waitlist_bans_manage", "Can Manage Bans"),  # Leadership
            ("waitlist_commanders_view", "can view commanders"),  # FC
            ("waitlist_commanders_manage", "can manage commanders"),  # Leadership
            ("waitlist_documentation_view", "can view documentation"),  # Jr FC
            ("waitlist_fleet_view", "Can View Active Fleet composition"),  # FC
            ("waitlist_implants_view", "Can View Implants"),  # FC
            ("waitlist_history_view", "Can View History"),  # Leadership
            ("waitlist_notes_view", "Can View Notes"),  # FC
            ("waitlist_notes_manage", "Can Manage notes"),  # Leadership
            ("waitlist_search", "Can Search"),  # FC
            ("waitlist_stats_view", "Can View Fleet Stats"),  # Leadership
            ("waitlist_skills_view", "Can view Skills"),  # Leadership
            # Add extra context to the Waitlist for Non-FCs
            ("waitlist_context_a", "Add Context to Waitlist: Number of Pilots"),
            ("waitlist_context_b", "Add Context to Waitlist: Ship Types"),
            ("waitlist_context_c", "Add Context to Waitlist: Time in Waitlist"),
            ("waitlist_context_d", "Add Context to Waitlist: Pilot Names"),

            ("waitlist_manage_waitlist", "Manage Waitlist"),  # FC
            ("waitlist_manage_waitlist_approve_fits", "Can Approve and Deny Fits on the WL"),  # FC

            # ESI Calls, Limit these maybe?
            ("waitlist_esi_show_info", "Can open a Show Info Window ingame"),
            ("waitlist_esi_search", "Will search ESI when using Search"),
        )



# Generated by Django 4.2.13 on 2024-07-05 05:24

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveonline', '0017_alliance_and_corp_names_are_not_unique'),
        ('authentication', '0023_alter_userprofile_language'),
        ('eveuniverse', '0010_alter_eveindustryactivityduration_eve_type_and_more'),
        ('incursions', '0007_alter_focus_options'),
    ]

    operations = [
        migrations.CreateModel(
            name='General',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'permissions': (('basic_incursions', 'Can view incursion data'),),
                'managed': False,
                'default_permissions': (),
            },
        ),
        migrations.RemoveField(
            model_name='incursionsconfig',
            name='security_high',
        ),
        migrations.RemoveField(
            model_name='incursionsconfig',
            name='security_low',
        ),
        migrations.RemoveField(
            model_name='incursionsconfig',
            name='security_null',
        ),
        migrations.AddField(
            model_name='webhook',
            name='security_high',
            field=models.BooleanField(default=True, verbose_name='Notify on High Security Incursions'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='webhook',
            name='security_low',
            field=models.BooleanField(default=True, verbose_name='Notify on Low Security Incursions'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='webhook',
            name='security_null',
            field=models.BooleanField(default=True, verbose_name='Notify on Null Security Incursions'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='incursion',
            name='faction',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveonline.evefactioninfo', verbose_name='The attacking Faction'),
        ),
        migrations.CreateModel(
            name='WaitlistShip',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('weight', models.IntegerField(verbose_name='Weight')),
                ('ship', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evetype', verbose_name='Ship Hull')),
            ],
            options={
                'verbose_name': 'Waitlist Ship Weight',
                'verbose_name_plural': 'Waitlist Ship Weightings',
            },
        ),
        migrations.CreateModel(
            name='WaitlistEntry',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('time', models.DateTimeField(auto_now_add=True, verbose_name='Waitlist Join Time')),
                ('note', models.TextField(verbose_name='Any extended note, Box #, begging to let ishtar-trash into fleet')),
                ('role', models.CharField(choices=[('dps', 'Mainline DPS'), ('vindicator', 'Vindicator'), ('logistics', 'Logistics'), ('boosts', 'Fleet Boosts / Other Utility')], default='dps', max_length=50, verbose_name='Ship role, usually applied to split the waitlist.')),
                ('ship', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='incursions.waitlistship', verbose_name='Ship Hull, with applied Weight')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='authentication.user', verbose_name='User')),
            ],
            options={
                'verbose_name': 'Waitlist Entry',
                'verbose_name_plural': 'Waitlist Entries',
            },
        ),
        migrations.CreateModel(
            name='Waitlist',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('open', models.DateTimeField(verbose_name='')),
                ('close', models.DateTimeField(verbose_name='')),
                ('fc', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='authentication.user', verbose_name='')),
            ],
            options={
                'verbose_name': 'Waitlist',
                'verbose_name_plural': 'Waitlist',
            },
        ),
    ]



from corptools.models import EveItemType
from corptools.task_helpers.update_tasks import process_bulk_types_from_esi
from ninja import NinjaAPI, Schema

from allianceauth.services.hooks import get_extension_logger

from incursions.models.waitlist import ApprovedFitting


class ModuleSchema(Schema):
    name: str
    category: str
    slot: str | None


class ModuleInfoRequest(Schema):
    ids: list[int]


logger = get_extension_logger(__name__)

api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    ModulesAPIEndpoints(api)


class ModulesAPIEndpoints:

    tags = ["Modules"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.post("/module/info", response={200: dict[int, ModuleSchema], 403: dict}, tags=self.tags)
        def module_info(request, payload: ModuleInfoRequest):
            if not request.user.has_perm("incursions.basic_waitlist"):
                return 403, {"error": "Permission denied"}

            process_bulk_types_from_esi(payload.ids)  # Make sure we have every TypeID to pass

            types = EveItemType.objects.prefetch_related("group__category").filter(type_id__in=payload.ids)

            return {
                type.type_id: ModuleSchema(
                    name=type.name,
                    category=type.group.category.name,
                    slot=None
                ) for type in types
            }

        @api.get("/module/preload", response={200: dict[int, ModuleSchema], 403: dict}, tags=self.tags)
        def preload(request):
            if not request.user.has_perm("incursions.basic_waitlist"):
                return 403, {"error": "Permission denied"}

            type_ids: list[int] = []
            for fit in ApprovedFitting.objects.all():
                for x in fit.dna.split(":"):
                    if x:
                        type_ids.append(int(x.split(";")[0]))

            process_bulk_types_from_esi(type_ids)  # Make sure we have every TypeID to pass
            types = EveItemType.objects.prefetch_related("group__category").filter(type_id__in=type_ids)
            return {
                type.type_id: ModuleSchema(
                    name=type.name,
                    category=type.group.category.name if type.group.category else None,
                    slot=None
                ) for type in types
            }



from corptools.models import EveItemType

from django.contrib import admin
from django.forms import ModelChoiceField

from allianceauth.services.hooks import get_extension_logger

from incursions.models.incursion import (
    Focus, Incursion, IncursionInfluence, IncursionsConfig, Webhook,
)
from incursions.models.waitlist import (
    ActiveFleet, Announcement, ApprovedFitting, ApprovedImplantSet,
    ApprovedSkills, Badge, Ban, CharacterBadges, CharacterNote, CharacterRoles,
    Fitting, FittingHistory, Fleet, FleetActivity, FleetSquad, ImplantSet,
    Role, SkillCheck, SkillHistory, Waitlist, WaitlistCategory,
    WaitlistCategoryRule, WaitlistEntry, WaitlistEntryFit,
)

logger = get_extension_logger(__name__)


EVECATEGORY_SHIP = 6
EVECATEGORY_SKILLS = 16


# Incursions
@admin.register(IncursionsConfig)
class IncursionsConfigAdmin(admin.ModelAdmin):
    filter_horizontal = ["status_webhooks", ]


@admin.register(Incursion)
class IncursionAdmin(admin.ModelAdmin):
    list_display = ["constellation", "state", "established_timestamp",
                    "mobilizing_timestamp", "withdrawing_timestamp", "ended_timestamp"]
    list_filter = ["state", "has_boss"]
    filter_horizontal = ["infested_solar_systems", ]


@admin.register(Webhook)
class WebhookAdmin(admin.ModelAdmin):
    list_display = ("name", "url")


@admin.register(IncursionInfluence)
class IncursionInfluenceAdmin(admin.ModelAdmin):
    list_display = ("incursion", "timestamp", "influence")


@admin.register(Focus)
class FocusAdmin(admin.ModelAdmin):
    list_display = ["incursion",]

    def formfield_for_foreignkey(self, db_field, request, **kwargs):
        if db_field.name == "incursion":
            kwargs["queryset"] = Incursion.objects.exclude(state=Incursion.States.ENDED)
        return super().formfield_for_foreignkey(db_field, request, **kwargs)


@admin.register(Role)
class RoleAdmin(admin.ModelAdmin):
    list_display = ("name", "description")
    search_fields = ("name", "description")


@admin.register(CharacterRoles)
class CharacterRolesAdmin(admin.ModelAdmin):
    list_display = ("character", "role", "granted_at", "granted_by")
    list_filter = ("role",)
    search_fields = ("character__name", "role__name")
    ordering = ("-granted_at",)


@admin.register(Announcement)
class AnnouncementAdmin(admin.ModelAdmin):
    list_display = ("message", "is_alert", "created_at", "revoked_at")
    list_filter = ("is_alert",)
    search_fields = ("message",)
    date_hierarchy = "created_at"


@admin.register(Ban)
class BanAdmin(admin.ModelAdmin):
    list_display = ("pk", "entity_name", "entity_type", "issued_at", "revoked_at")
    list_filter = ("entity_type", "issued_at", "revoked_at")
    search_fields = ("entity_name", "public_reason", "reason")


@admin.register(Badge)
class BadgeAdmin(admin.ModelAdmin):
    list_display = ("name", "exclude_badge")
    search_fields = ("name",)


@admin.register(CharacterBadges)
class CharacterBadgesAdmin(admin.ModelAdmin):
    list_display = ("pk", "character", "badge", "granted_at", "granted_by")
    list_filter = ("badge",)
    search_fields = ("character__name", "badge__name")


@admin.register(Fitting)
class FittingAdmin(admin.ModelAdmin):
    list_display = ("pk", "ship", "dna")
    search_fields = ("dna", "ship__name")


@admin.register(ImplantSet)
class ImplantSetAdmin(admin.ModelAdmin):
    list_display = ("pk", "implants")
    search_fields = ("implants",)


@admin.register(FittingHistory)
class FittingHistoryAdmin(admin.ModelAdmin):
    list_display = ("pk", "character", "fit", "implant_set", "logged_at")
    list_filter = ("fit", "implant_set")
    search_fields = ("character__name",)


@admin.register(FleetActivity)
class FleetActivityAdmin(admin.ModelAdmin):
    list_display = (
        "character",
        "ship",
        "fleet_id",
        "has_left",
        "is_boss",
        "first_seen",
        "last_seen",
    )
    list_filter = ("has_left", "is_boss")
    search_fields = ("character__name", "ship__name")


@admin.register(SkillHistory)
class SkillHistoryAdmin(admin.ModelAdmin):
    list_display = ("pk", "character", "skill", "old_level", "new_level", "logged_at")
    search_fields = ("character__name", "skill__name")


@admin.register(CharacterNote)
class CharacterNoteAdmin(admin.ModelAdmin):
    list_display = ("character", "author", "logged_at")
    search_fields = ("character__name", "author__name", "note")


@admin.register(WaitlistCategory)
class WaitlistCategoryAdmin(admin.ModelAdmin):
    list_display = ("name", )
    search_fields = ("name",)


@admin.register(WaitlistCategoryRule)
class WaitlistCategoryRuleAdmin(admin.ModelAdmin):

    def formfield_for_foreignkey(self, db_field, request, **kwargs) -> ModelChoiceField:
        if db_field.name == "ship":
            kwargs["queryset"] = EveItemType.objects.filter(group__category_id=EVECATEGORY_SHIP)

        return super().formfield_for_foreignkey(db_field, request, **kwargs)


@admin.register(Fleet)
class FleetAdmin(admin.ModelAdmin):
    list_display = ("pk", "boss", "is_updating")
    search_fields = ("boss__name",)


@admin.register(FleetSquad)
class FleetSquadAdmin(admin.ModelAdmin):
    list_display = ("pk", "fleet", "category", "wing_id", "squad_id")
    search_fields = ("fleet__id", "category__name")


@admin.register(Waitlist)
class WaitlistAdmin(admin.ModelAdmin):
    list_display = ("pk", "name", "is_open")
    list_filter = ("is_open", )
    search_fields = ("name",)


@admin.register(WaitlistEntryFit)
class WaitlistEntryFitAdmin(admin.ModelAdmin):
    list_display = (
        "pk",
        "character",
        "fit",
        "implant_set",
        "approved",
        "tags",
        "category",
        "cached_time_in_fleet",
        "is_alt",
    )
    list_filter = ("approved", "is_alt", "fit", "implant_set", "category")
    search_fields = ("character__name", "tags", "category__name")


@admin.register(ApprovedFitting)
class ApprovedFittingAdmin(admin.ModelAdmin):
    list_display = ("pk", "ship", "dna")
    search_fields = ("dna", "ship__name")


@admin.register(ApprovedImplantSet)
class ApprovedImplantSetAdmin(admin.ModelAdmin):
    list_display = ("pk", "implants")
    search_fields = ("implants",)


@admin.register(ActiveFleet)
class ActiveFleetAdmin(admin.ModelAdmin):
    list_display = ("pk", "fleet")
    search_fields = ("fleet__id", "fleet__boss__character_name")


@admin.register(SkillCheck)
class SkillCheckAdmin(admin.ModelAdmin):
    list_display = ("skill", "min", "elite", "gold")
    search_fields = ("skill__name",)
    ordering = ("skill",)

    def formfield_for_foreignkey(self, db_field, request, **kwargs) -> ModelChoiceField:
        if db_field.name == "skill":
            kwargs["queryset"] = EveItemType.objects.filter(group__category_id=EVECATEGORY_SKILLS)

        return super().formfield_for_foreignkey(db_field, request, **kwargs)


@admin.register(ApprovedSkills)
class ApprovedSkillsAdmin(admin.ModelAdmin):
    list_display = ("hull",)
    search_fields = ("hull__name",)
    filter_horizontal = ("skill_checks",)

    def formfield_for_foreignkey(self, db_field, request, **kwargs) -> ModelChoiceField:
        if db_field.name == "hull":
            kwargs["queryset"] = EveItemType.objects.filter(group__category_id=EVECATEGORY_SHIP)

        return super().formfield_for_foreignkey(db_field, request, **kwargs)


@admin.register(WaitlistEntry)
class WaitlistEntryAdmin(admin.ModelAdmin):
    list_display = ("pk", "main_character", "joined_at", "waitlist")
    search_fields = ("main_character__character_name",)
    list_filter = ("waitlist",)
    date_hierarchy = "joined_at"



import msgpack
from ninja import NinjaAPI

from django.shortcuts import redirect

from allianceauth.framework.api.user import get_main_character_from_user
from allianceauth.services.hooks import get_extension_logger

from incursions import app_settings
from incursions.api.helpers.sse import SSEClient

logger = get_extension_logger(__name__)
api = NinjaAPI()

VERSION_BYTE = b'\xBA'


def setup(api: NinjaAPI) -> None:
    SSEAPIEndpoints(api)


class SSEAPIEndpoints:

    tags = ["SSE"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/sse/stream", tags=self.tags)
        def sse_stream(request):
            if not request.user.has_perm("incursions.basic_waitlist"):
                return {"error": "Permission denied"}

            topics = ["announcements", "waitlist", f"account;{get_main_character_from_user(request.user).character_id}"]
            if request.user.is_authenticated and request.user.has_perm("waitlist_fleet_view"):
                topics.append("fleet_comp")

            payload_dict = {"topics": topics}
            payload_bytes: bytes = msgpack.packb(payload_dict, use_bin_type=True)

            token = SSEClient(app_settings.SSE_SITE_URL, app_settings.SSE_SECRET).branca_encode(payload=payload_bytes)

            sse_url = f"{app_settings.SSE_SITE_URL}/events?token={token}"
            return redirect(sse_url)



from datetime import datetime, timezone

from ninja import NinjaAPI, Schema

from django.shortcuts import get_object_or_404

from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_main_character_from_user
from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import (
    AllianceSchema, CharacterSchema, CorporationSchema,
)
from incursions.models.waitlist import Ban


class PublicBanSchema(Schema):
    id: int
    entity_name: str
    entity_type: str
    entity_character: CharacterSchema | None
    entity_corporation: CorporationSchema | None
    entity_alliance: AllianceSchema | None
    issued_at: datetime
    issued_by: CharacterSchema | None
    public_reason: str | None
    revoked_at: datetime | None
    revoked_by: CharacterSchema | None


class BanSchema(Schema):
    id: int
    entity_name: str
    entity_type: str
    entity_character: CharacterSchema | None
    entity_corporation: CorporationSchema | None
    entity_alliance: AllianceSchema | None
    issued_at: datetime
    issued_by: CharacterSchema | None
    reason: str
    public_reason: str | None
    revoked_at: datetime | None
    revoked_by: CharacterSchema | None


class UpdateBanSchema(Schema):
    id: int
    entity_type: str
    entity_character: CharacterSchema | None
    entity_corporation: CorporationSchema | None
    entity_alliance: AllianceSchema | None
    reason: str
    public_reason: str | None
    revoked_at: datetime | None


logger = get_extension_logger(__name__)

api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    BansAPIEndpoints(api)


class BansAPIEndpoints:

    tags = ["Bans"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/bans", response={200: list[BanSchema], 403: dict}, tags=self.tags)
        def list_bans(request):
            if not request.user.has_perm("incursions.view_bans"):
                return 403, {"error": "Permission denied"}

            return 200, Ban.objects.filter(revoked_at__isnull=True)

        @api.post("/bans", response={200: dict, 403: dict}, tags=self.tags)
        def create_ban(request, payload: UpdateBanSchema):
            if not request.user.has_perm("incursions.manage_bans"):
                return 403, {"error": "Permission denied"}

            Ban.objects.create(
                pk=payload.id,
                entity_type=payload.entity_type,
                issued_at=datetime.now(timezone.utc),
                issued_by=get_main_character_from_user(request.user),
                entity_character=payload.entity_character,
                entity_corporation=payload.entity_corporation,
                entity_alliance=payload.entity_alliance,
                reason=payload.reason,
                public_reason=payload.public_reason,
                revoked_at=payload.revoked_at,
                revoked_by=get_main_character_from_user(request.user) if payload.revoked_at is not None else None
            )
            return 200, {"status": "Ban created"}

        @api.get("/bans/{character_id}", response={200: list[BanSchema], 403: dict, 404: dict}, tags=self.tags)
        def character_history(request, character_id: int):
            if not request.user.has_perm("incursions.view_bans"):
                return 403, {"error": "Permission denied"}

            try:
                character = get_object_or_404(EveCharacter, character_id=character_id)
                bans = Ban.objects.filter(entity_character=character, entity_type="Character")
            except EveCharacter.DoesNotExist:
                return 404, {"error": "Character not found"}
            except Ban.DoesNotExist:
                return 404, {"error": "Ban not found"}

            return [BanSchema.from_orm(ban) for ban in bans]

        @api.patch("/bans/{ban_id}", response={200: dict, 403: dict, 404: dict}, tags=self.tags)
        def update_ban(request, ban_id: int, payload: BanSchema):
            if not request.user.has_perm("incursions.manage_bans"):
                return 403, {"error": "Permission denied"}

            try:
                ban = Ban.objects.get(id=ban_id)
            except Ban.DoesNotExist:
                return 404, {"error": "Ban not found"}

            ban.reason = payload.reason
            ban.public_reason = payload.public_reason
            ban.revoked_at = payload.revoked_at
            ban.revoked_by = get_main_character_from_user(request.user) if payload.revoked_at is not None else None
            ban.issued_by = get_main_character_from_user(request.user)
            ban.issued_at = datetime.now(timezone.utc)
            ban.save()

            return 200, {"status": "Ban updated"}

        @api.delete("/bans/{ban_id}", response={200: dict, 403: dict, 404: dict}, tags=self.tags)
        def revoke_ban(request, ban_id: int)  :
            if not request.user.has_perm("incursions.manage_bans"):
                return 403, {"error": "Permission denied"}

            try:
                ban = Ban.objects.get(id=ban_id)
            except Ban.DoesNotExist:
                return 404, {"error": "Ban not found"}

            ban.revoked_at = datetime.now(timezone.utc)
            ban.revoked_by = get_main_character_from_user(request.user)
            ban.save()
            return 200, {"status": "Ban revoked"}



from ninja import NinjaAPI
from ninja.security import django_auth

from django.conf import settings

from allianceauth.services.hooks import get_extension_logger

from incursions.api import (
    announcements, auth, badges, bans, categories, commanders, fittings,
    fleets, implants, modules, notes, pilot, search, skills, sse, waitlist,
    window,
)

logger = get_extension_logger(__name__)

api = NinjaAPI(
    title="AA Incursions API",
    version="0.0.1",
    urls_namespace='incursions:api',
    auth=django_auth,
    csrf=True,
    openapi_url=settings.DEBUG and "/openapi.json" or ""
)

announcements.setup(api)
auth.setup(api)
badges.setup(api)
bans.setup(api)
categories.setup(api)
commanders.setup(api)
fittings.setup(api)
fleets.setup(api)
# history.setup(api)
implants.setup(api)
modules.setup(api)
notes.setup(api)
pilot.setup(api)
search.setup(api)
# skillplans.setup(api)
skills.setup(api)
sse.setup(api)
# statistics.setup(api)
waitlist.setup(api)
window.setup(api)



import re

from corptools.models import EveItemType

# Always Cargo Categories
EVECATEGORY_IMPLANTS = 20
EVECATEGORY_CHARGE = 8
EVECATEGORY_DEPLOYABLE = 22

ALWAYS_CARGO_CATEGORIES = [EVECATEGORY_IMPLANTS, EVECATEGORY_CHARGE, EVECATEGORY_DEPLOYABLE]


class FitError(Exception):
    pass


class FittingParser:
    def __init__(self, hull: int, modules: dict[int, int] | None = None, cargo: dict[int, int] | None = None):
        self.hull = hull
        self.modules = modules if modules is not None else {}
        self.cargo = cargo if cargo is not None else {}

    @classmethod
    def from_dna(cls, dna: str) -> "FittingParser":
        pieces = dna.split(':')
        if not pieces or pieces[0] == "":
            raise FitError("DNA string is empty or missing hull")
        try:
            hull = int(pieces[0])
        except ValueError:
            raise FitError("Invalid hull type ID")
        cargo: dict[int, int] = {}
        modules: dict[int, int] = {}
        i = 0
        for piece in pieces[1:]:
            if not piece:
                continue
            i += 1
            if i > 1000:
                raise FitError("Too many pieces in DNA")
            mod_split = piece.split(';', 1)
            type_id_str = mod_split[0]
            # If the type id ends with '_' then it is cargo
            if type_id_str.endswith('_'):
                try:
                    type_id = int(type_id_str[:-1])
                except ValueError:
                    raise FitError("Invalid type ID in cargo")
                is_cargo = True
            else:
                try:
                    type_id = int(type_id_str)
                except ValueError:
                    raise FitError("Invalid type ID")
                # Use your database to check if this type is always cargo.
                is_cargo = True if EveItemType.objects.get(type_id=type_id).group.category.category_id in ALWAYS_CARGO_CATEGORIES else False
            # Count defaults to 1 if missing.
            count = 1
            if len(mod_split) == 2 and mod_split[1]:
                try:
                    count = int(mod_split[1])
                except ValueError:
                    raise FitError("Invalid count in DNA")
            # Choose dictionary based on whether this is cargo or module.
            dest = cargo if is_cargo else modules
            dest[type_id] = dest.get(type_id, 0) + count
        return cls(hull, modules, cargo)

    def to_dna(self) -> str:
        # Start with hull followed by a colon
        dna = f"{self.hull}:"
        # Append modules: format "{id};{count}:"
        for type_id, count in self.modules.items():
            dna += f"{type_id};{count}:"
        # Append cargo: if type is not always cargo, add an underscore after the id.
        for type_id, count in self.cargo.items():
            if EveItemType.objects.get(type_id=type_id).group.category.category_id in ALWAYS_CARGO_CATEGORIES:
                dna += f"{type_id};{count}:"
            else:
                dna += f"{type_id}_;{count}:"
        return dna + ":"

    @classmethod
    def from_eft(cls, eft: str) -> "FittingParser":
        fit: FittingParser | None = None
        section = 0

        for line in eft.splitlines():
            line = line.strip()

            # Look for the fitting header.
            if line.startswith('[') and line.endswith(']') and ',' in line:
                if fit is not None:
                    raise FitError("Multiple fitting headers found; expected only one fitting in EFT input.")
                # Remove the brackets and split on the first comma.
                line_inner = line[1:-1]
                pieces = line_inner.split(',', 1)
                hull_name = pieces[0].strip()
                if not pieces[1].strip():
                    raise FitError("Missing ship name in EFT header")
                # Look up the hull ID from the name.
                hull = EveItemType.objects.get(name__iexact=hull_name).type_id
                fit = cls(hull, modules={}, cargo={})
                section = 0  # Reset section counter for the current fit.
            elif fit is not None:
                # If the line indicates an empty section, increase section counter.
                if line == "":
                    section += 1
                    continue

                # Skip lines that indicate empty modules.
                if line.startswith("[Empty "):
                    continue

                # Parse a module line.
                m = re.match(r"(.+?)(?:\s*x\s*(\d+))?$", line)
                if not m:
                    raise FitError("Could not parse EFT line: " + line)
                type_name = m.group(1).strip()
                try:
                    type_id = EveItemType.objects.get(name__iexact=type_name).type_id
                except Exception as e:
                    raise FitError(f"Error looking up type ID for '{type_name}': {e}")
                count = int(m.group(2)) if m.group(2) is not None else 1
                stacked = m.group(2) is not None

                # Determine if this item is cargo:
                # If section >= 7 then treat as cargo; otherwise, query the type info.
                if section >= 7:
                    is_cargo = True
                else:
                    type_obj = EveItemType.objects.get(type_id=type_id)

                    is_cargo = type_obj.group.category.category_id in ALWAYS_CARGO_CATEGORIES or (stacked and type_obj.group.category.name != "Drone")

                # Add to cargo or modules.
                dest = fit.cargo if is_cargo else fit.modules
                dest[type_id] = dest.get(type_id, 0) + count
            else:
                # Skip lines until a header is found.
                continue

        if fit is None:
            raise FitError("Invalid EFT format; no fitting header found")

        return fit



from corptools.models import EveItemType
from ninja import NinjaAPI, Schema

from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import CharacterSchema, HullSchema
from incursions.api.skills import EVECATEGORY_SKILLS, RELEVANT_SKILL_GROUPS
from incursions.models.waitlist import (
    FittingHistory, FleetActivity, SkillHistory,
)

api = NinjaAPI()


class ActivityEntrySchema(Schema):
    hull: HullSchema
    logged_at: int
    time_in_fleet: int


class ActivitySummaryEntrySchema(Schema):
    hull: HullSchema
    time_in_fleet: int


class ActivityResponse(Schema):
    activity: list[ActivityEntrySchema]
    summary: list[ActivitySummaryEntrySchema]


class FleetCompEntrySchema(Schema):
    hull: HullSchema
    character: CharacterSchema
    logged_at: int
    time_in_fleet: int
    is_boss: bool


class FleetCompResponse(Schema):
    fleets: dict[int, list[FleetCompEntrySchema]]


class XupHistoryLineSchema(Schema):
    logged_at: int
    dna: str
    implants: list[int]
    hull: HullSchema


class XupHistorySchema(Schema):
    xups: list[XupHistoryLineSchema]


class SkillHistoryResponseLineSchema(Schema):
    skill_id: int
    old_level: int
    new_level: int
    logged_at: int
    name: str


class SkillHistoryResponse(Schema):
    history: list[SkillHistoryResponseLineSchema]
    ids: dict[str, int]


logger = get_extension_logger(__name__)


def setup(api: NinjaAPI) -> None:
    HistoryAPIEndpoints(api)


class HistoryAPIEndpoints:

    tags = ["History"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/api/history/fleet", response=ActivityResponse, tags=self.tags)
        def fleet_history(request, character_id: int):
            if not request.user.has_perm("incursions.waitlist_history_view"):
                return 403, {"error": "Permission denied"}

            activities = FleetActivity.objects.filter(character_id=character_id).order_by("-first_seen")
            time_by_hull: dict[int, int] = {}
            activity_entries: list[ActivityEntrySchema] = []

            for act in activities:
                time_in_fleet = act.last_seen - act.first_seen
                hull_id = act.ship.type_id  # Assuming "hull" is an integer field on FleetActivity.
                time_by_hull[hull_id] = time_by_hull.get(hull_id, 0) + time_in_fleet

                activity_entries.append(ActivityEntrySchema(
                    hull=HullSchema(id=hull_id, name=EveItemType.objects.get(id=hull_id).name),
                    logged_at=act.first_seen,
                    time_in_fleet=time_in_fleet,
                ))

            summary_entries: list[ActivitySummaryEntrySchema] = []
            for hull_id, total_time in time_by_hull.items():
                summary_entries.append(ActivitySummaryEntrySchema(
                    hull=HullSchema(id=hull_id, name=EveItemType.objects.get(id=hull_id).name),
                    time_in_fleet=total_time,
                ))
            summary_entries.sort(key=lambda s: s.time_in_fleet, reverse=True)

            return {
                "activity": activity_entries,
                "summary": summary_entries,
            }

        @api.get("/api/history/fleet-comp", response=FleetCompResponse, tags=self.tags)
        def fleet_comp(request, time: int):
            if not request.user.has_perm("incursions.waitlist_history_view"):
                return 403, {"error": "Permission denied"}

            comp_entries = FleetActivity.objects.select_related("character").filter(
                first_seen__lte=time,
                last_seen__gte=time,
            )

            fleets: dict[int, list[FleetCompEntrySchema]] = {}
            for entry in comp_entries:
                fleet_id = entry.fleet_id  # Adjust field name if necessary.
                fleet_entry = FleetCompEntrySchema(
                    hull=HullSchema.from_orm(entry.ship),
                    character=CharacterSchema.from_orm(entry.character),
                    logged_at=entry.first_seen,
                    time_in_fleet=entry.last_seen - entry.first_seen,
                    is_boss=True if entry.fleet.boss == entry.character else False,
                )
                fleets.setdefault(fleet_id, []).append(fleet_entry)

            return {"fleets": fleets}

        @api.get("/api/history/xup", response=XupHistorySchema, tags=self.tags)
        def xup_history(request, character_id: int):
            if not request.user.has_perm("incursions.waitlist_history_view"):
                return 403, {"error": "Permission denied"}

            xup_lines: list[XupHistoryLineSchema] = []
            for xup in FittingHistory.objects.filter(character_id=character_id).order_by("-id"):
                try:
                    implants_list = [int(i) for i in xup.implant_set.implants.split(':') if i]
                except Exception:
                    implants_list: list[int] = []
                hull_schema = HullSchema().from_orm(xup.fit.ship)
                xup_lines.append(XupHistoryLineSchema(
                    logged_at=xup.logged_at,
                    dna=xup.fit.dna,
                    implants=implants_list,
                    hull=hull_schema,
                ))

            return {"xups": xup_lines}

        @api.get("/api/history/skills", response=SkillHistoryResponse, tags=self.tags)
        def skill_history(request, character_id: int):
            if not request.user.has_perm("incursions.waitlist_history_view"):
                return 403, {"error": "Permission denied"}

            history_list: list[SkillHistoryResponseLineSchema] = []
            for skill in SkillHistory.objects.filter(character_id=character_id).order_by("-id"):
                if skill.skill.group.category.category_id in RELEVANT_SKILL_GROUPS:
                    history_list.append(SkillHistoryResponseLineSchema(
                        skill_id=skill.skill.type_id,
                        old_level=skill.old_level,
                        new_level=skill.new_level,
                        logged_at=skill.logged_at,
                        name=skill.skill.name,
                    ))

            return {
                "history": history_list,
                "ids": {skill.name: skill.type_id for skill in EveItemType.objects.filter(group__category_id=EVECATEGORY_SKILLS)},
            }



from ninja import Schema


class CharacterSchema(Schema):
    character_id: int
    character_name: str
    corporation_id: int | None
    alliance_id: int | None


class CorporationSchema(Schema):
    corporation_id: int
    corporation_name: str
    alliance_id: int | None


class AllianceSchema(Schema):
    alliance_id: int
    alliance_name: str


class HullSchema(Schema):
    id: int
    name: str



# Generated by Django 4.2.19 on 2025-03-29 08:40

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    replaces = [('incursions', '0001_initial'), ('incursions', '0002_remove_incursion_influence_and_more'), ('incursions', '0003_focus'), ('incursions', '0004_alter_focus_incursion'), ('incursions', '0005_incursioninfluence_uniqueincursioninfluencelogtimestamp'), ('incursions', '0006_alter_incursion_infested_solar_systems_and_more'), ('incursions', '0007_alter_focus_options'), ('incursions', '0008_general_remove_incursionsconfig_security_high_and_more'), ('incursions', '0009_alter_incursion_state_alter_waitlist_close_and_more')]

    initial = True

    dependencies = [
        ('eveonline', '0017_alliance_and_corp_names_are_not_unique'),
        ('eveuniverse', '0007_evetype_description'),
        ('eveuniverse', '0010_alter_eveindustryactivityduration_eve_type_and_more'),
        ('authentication', '0023_alter_userprofile_language'),
    ]

    operations = [
        migrations.CreateModel(
            name='Webhook',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('url', models.CharField(max_length=200)),
                ('name', models.CharField(max_length=50)),
            ],
            options={
                'verbose_name': 'Destination Webhook',
                'verbose_name_plural': 'Destination Webhooks',
            },
        ),
        migrations.CreateModel(
            name='IncursionsConfig',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('security_high', models.BooleanField(verbose_name='Notify on High Security Incursions')),
                ('security_low', models.BooleanField(verbose_name='Notify on Low Security Incursions')),
                ('security_null', models.BooleanField(verbose_name='Notify on Null Security Incursions')),
                ('status_webhooks', models.ManyToManyField(to='incursions.webhook', verbose_name='Destination Webhook for Incursion Updates')),
            ],
            options={
                'verbose_name': 'AA Incursions Settings',
                'verbose_name_plural': 'AA Incursions Settings',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Incursion',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('has_boss', models.BooleanField(default=False, verbose_name='Whether the final encounter has boss or not')),
                ('state', models.CharField(choices=[('established', 'Incursion established'), ('mobilizing', 'Incursion mobilized'), ('withdrawing', 'Incursion withdrawing'), ('ended', 'Ended')], default='established', max_length=50, verbose_name='The state of this incursion')),
                ('type', models.CharField(choices=[('Incursion', 'Sansha Incursion')], default='Incursion', max_length=50, verbose_name='The type of this incursion')),
                ('established_timestamp', models.DateTimeField(blank=True, null=True, verbose_name='Established')),
                ('mobilizing_timestamp', models.DateTimeField(blank=True, null=True, verbose_name='Mobilizing')),
                ('withdrawing_timestamp', models.DateTimeField(blank=True, null=True, verbose_name='Withdrawing')),
                ('ended_timestamp', models.DateTimeField(blank=True, null=True, verbose_name='Ended')),
                ('constellation', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.eveconstellation', verbose_name='The constellation id in which this incursion takes place')),
                ('faction', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveonline.evefactioninfo', verbose_name='The attacking faction’s id')),
                ('infested_solar_systems', models.ManyToManyField(related_name='+', to='eveuniverse.evesolarsystem', verbose_name='A list of infested solar system ids that are a part of this incursion')),
                ('staging_solar_system', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveuniverse.evesolarsystem', verbose_name='Staging solar system for this incursion')),
            ],
            options={
                'verbose_name': 'Incursion',
                'verbose_name_plural': 'Incursions',
            },
        ),
        migrations.CreateModel(
            name='IncursionInfluence',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('timestamp', models.DateTimeField(verbose_name='Timestamp')),
                ('influence', models.FloatField(default=0.0, verbose_name='Influence of this incursion as a float from 0 to 1')),
                ('incursion', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='incursions.incursion', verbose_name='Incursion')),
            ],
            options={
                'verbose_name': 'IncursionInfluence',
                'verbose_name_plural': 'IncursionInfluences',
            },
        ),
        migrations.CreateModel(
            name='Focus',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('incursion', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='incursions.incursion', verbose_name='Current Focus')),
            ],
            options={
                'verbose_name': 'Focus',
                'verbose_name_plural': 'Focus',
                'default_permissions': (),
            },
        ),
        migrations.AddConstraint(
            model_name='incursioninfluence',
            constraint=models.UniqueConstraint(fields=('incursion', 'timestamp'), name='UniqueIncursionInfluenceLogTimestamp'),
        ),
        migrations.AlterField(
            model_name='incursion',
            name='infested_solar_systems',
            field=models.ManyToManyField(related_name='+', to='eveuniverse.evesolarsystem', verbose_name='A list of infested solar system ids that are a part of this incursion'),
        ),
        migrations.AlterField(
            model_name='incursionsconfig',
            name='status_webhooks',
            field=models.ManyToManyField(to='incursions.webhook', verbose_name='Destination Webhook for Incursion Updates'),
        ),
        migrations.AlterModelOptions(
            name='focus',
            options={'default_permissions': (), 'verbose_name': 'Current Focus', 'verbose_name_plural': 'Current Focus'},
        ),
        migrations.CreateModel(
            name='General',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'permissions': (('basic_incursions', 'Can view incursion data'),),
                'managed': False,
                'default_permissions': (),
            },
        ),
        migrations.RemoveField(
            model_name='incursionsconfig',
            name='security_high',
        ),
        migrations.RemoveField(
            model_name='incursionsconfig',
            name='security_low',
        ),
        migrations.RemoveField(
            model_name='incursionsconfig',
            name='security_null',
        ),
        migrations.AddField(
            model_name='webhook',
            name='security_high',
            field=models.BooleanField(default=True, verbose_name='Notify on High Security Incursions'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='webhook',
            name='security_low',
            field=models.BooleanField(default=True, verbose_name='Notify on Low Security Incursions'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='webhook',
            name='security_null',
            field=models.BooleanField(default=True, verbose_name='Notify on Null Security Incursions'),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='incursion',
            name='faction',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveonline.evefactioninfo', verbose_name='The attacking Faction'),
        ),
        migrations.CreateModel(
            name='WaitlistShip',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('weight', models.IntegerField(verbose_name='Weight')),
                ('ship', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='eveuniverse.evetype', verbose_name='Ship Hull')),
            ],
            options={
                'verbose_name': 'Waitlist Ship Weight',
                'verbose_name_plural': 'Waitlist Ship Weightings',
            },
        ),
        migrations.CreateModel(
            name='WaitlistEntry',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('time', models.DateTimeField(auto_now_add=True, verbose_name='Waitlist Join Time')),
                ('note', models.TextField(verbose_name='Any extended note, Box #, begging to let ishtar-trash into fleet')),
                ('role', models.CharField(choices=[('dps', 'Mainline DPS'), ('vindicator', 'Vindicator'), ('logistics', 'Logistics'), ('boosts', 'Fleet Boosts / Other Utility')], default='dps', max_length=50, verbose_name='Ship role, usually applied to split the waitlist.')),
                ('ship', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='incursions.waitlistship', verbose_name='Ship Hull, with applied Weight')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='authentication.user', verbose_name='User')),
            ],
            options={
                'verbose_name': 'Waitlist Entry',
                'verbose_name_plural': 'Waitlist Entries',
            },
        ),
        migrations.AlterField(
            model_name='incursion',
            name='state',
            field=models.CharField(choices=[('established', 'Incursion established'), ('mobilizing', 'Incursion mobilizing'), ('withdrawing', 'Incursion withdrawing'), ('ended', 'Ended')], default='established', max_length=50, verbose_name='The state of this incursion'),
        ),
        migrations.CreateModel(
            name='Waitlist',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('open', models.DateTimeField(verbose_name='Open Time')),
                ('close', models.DateTimeField(verbose_name='Close Time')),
                ('fc', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='authentication.user', verbose_name='Fleet Commander')),
            ],
            options={
                'verbose_name': 'Waitlist',
                'verbose_name_plural': 'Waitlist',
            },
        ),
    ]



from datetime import datetime

from ninja import NinjaAPI, Schema

from allianceauth.authentication.admin import User
from allianceauth.eveonline.models import EveCharacter
from allianceauth.framework.api.user import get_main_character_from_user
from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import CharacterSchema
from incursions.models.waitlist import Badge, CharacterBadges


class BadgeSchema(Schema):
    id: int
    name: str
    member_count: int
    exclude_badge_id: int | None


class CharacterBadgesSchema(Schema):
    badge: BadgeSchema
    character: CharacterSchema
    granted_by: CharacterSchema | None
    granted_at: datetime


logger = get_extension_logger(__name__)

api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    BadgesAPIEndpoints(api)


class BadgesAPIEndpoints:

    tags = ["Badges"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/badges", response={200: list[BadgeSchema], 403: dict}, tags=self.tags)
        def list_badges(request):
            if not request.user.has_perm("incursions.basic_waitlist"):
                return 403, {"error": "Permission denied"}

            return 200, Badge.objects.all()

        @api.get("/badges/{badge_id}/members", response={200: list[CharacterBadgesSchema], 403: dict, 404: dict}, tags=self.tags)
        def get_badge_members(request, badge_id: int):
            if not request.user.has_perm("incursions.manage_badges"):
                return 403, {"error": "Permission denied"}

            try:
                characters = CharacterBadges.objects.filter(badge_id=badge_id).select_related("character", "granted_by")
            except Badge.DoesNotExist:
                return 404, {"error": "Badge not found"}

            return 200, characters

        @api.post("/badges/{badge_id}/members", response={200: dict, 403: dict, 404: dict} ,tags=self.tags)
        def assign_badge(request, badge_id: int, character_id: int):
            if not request.user.has_perm("incursions.manage_badges"):
                return 403, {"error": "Permission denied"}

            try:
                character = EveCharacter.objects.filter(character_id=character_id).get()
            except EveCharacter.DoesNotExist:
                return 404, {"error": "EveCharacter not found"}

            try:
                badge = Badge.objects.filter(id=badge_id).get()
            except Badge.DoesNotExist:
                return 404, {"error": "Badge not found"}

            if CharacterBadges.objects.filter(character=character, badge=badge).exists():
                return 400, {"error": "EveCharacter already has this badge"}

            CharacterBadges.objects.create(badge=badge, character=character, granted_by=get_main_character_from_user(request.user))
            return 200, {"status": "Badge assigned"}

        @api.delete("/badges/{badge_id}/members/{character_id}", response={200: dict, 403: dict, 404: dict}, tags=self.tags)
        def revoke_badge(request, badge_id: int, character_id: int):
            if not request.user.has_perm("incursions.manage_badges"):
                return 403, {"error": "Permission denied"}

            try:
                character = EveCharacter.objects.filter(character_id=character_id).get()
            except User.DoesNotExist:
                return 404, {"error": "User not found"}

            try:
                badge = Badge.objects.filter(id=badge_id).get()
            except Badge.DoesNotExist:
                return 404, {"error": "Badge not found"}

            if not CharacterBadges.objects.filter(character=character, badge=badge).exists():
                return 404, {"error": "Badge assignment not found"}

            deleted, _ = CharacterBadges.objects.filter(character=character, badge_id=badge_id).delete()
            return 200, {"status": "Badge revoked"}



from django.contrib.auth.decorators import login_required
from django.http import HttpResponse
from django.shortcuts import redirect, render

from allianceauth.services.hooks import get_extension_logger
from esi.decorators import token_required

from incursions import __version__
from incursions.app_settings import INCURSIONS_SCOPES_BASE

logger = get_extension_logger(__name__)


@login_required
def waitlist(request) -> HttpResponse:
    return render(request, 'incursions/base-bs5.html', context={"version": __version__, "app_name": "incursions/waitlist", "page_title": "Waitlist"})


@login_required
@token_required(scopes=INCURSIONS_SCOPES_BASE)
def add_char(request, token):

    return redirect('incursions:waitlist')



# Generated by Django 4.2.19 on 2025-03-29 08:47

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eveonline', '0017_alliance_and_corp_names_are_not_unique'),
        ('corptools', '0118_alter_skill_id'),
        ('incursions', '0010_squashed_0009_alter_incursion_state_alter_waitlist_close_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='ActiveFleet',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
            ],
            options={
                'verbose_name': 'Waitlist - Fleet (Active)',
                'verbose_name_plural': 'Waitlist - Fleet (Active)',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Announcement',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('message', models.TextField()),
                ('is_alert', models.BooleanField(default=False)),
                ('pages', models.TextField(blank=True, null=True)),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('revoked_at', models.DateTimeField(blank=True, default=None, null=True)),
                ('created_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='eveonline.evecharacter')),
                ('revoked_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='eveonline.evecharacter')),
            ],
            options={
                'verbose_name': 'Waitlist - Announcement',
                'verbose_name_plural': 'Waitlist - Announcements',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='ApprovedFitting',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('dna', models.TextField(max_length=1024, unique=True)),
                ('tier', models.CharField(choices=[('Basic', 'Sponge'), ('Mainline', 'Mainline'), ('Alt', 'Alt'), ('Other', 'Other')], default='Mainline', max_length=16)),
                ('description', models.CharField(blank=True, max_length=255, null=True)),
            ],
            options={
                'verbose_name': 'Waitlist - Approved Fitting',
                'verbose_name_plural': 'Waitlist - Approved Fittings',
            },
        ),
        migrations.CreateModel(
            name='ApprovedImplantSet',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255)),
                ('set', models.CharField(choices=[('Amulet', 'Amulet'), ('Ascendency', 'Ascendency'), ('None', 'None')], default='Amulet', max_length=16)),
                ('implants', models.CharField(max_length=255, unique=True)),
            ],
            options={
                'verbose_name': 'Waitlist - Approved Implant Set',
                'verbose_name_plural': 'Waitlist - Approved Implant Sets',
            },
        ),
        migrations.CreateModel(
            name='ApprovedSkills',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('hull', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='corptools.eveitemtype', verbose_name='Hull')),
            ],
            options={
                'verbose_name': 'Waitlist - Approved Skills',
                'verbose_name_plural': 'Waitlist - Approved Skills',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Badge',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64, unique=True)),
                ('exclude_badge', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='excluded_by', to='incursions.badge')),
            ],
            options={
                'verbose_name': 'Waitlist - Badge',
                'verbose_name_plural': 'Waitlist - Badges',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Ban',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('entity_type', models.CharField(choices=[('Character', 'Character'), ('Corporation', 'Corporation'), ('Alliance', 'Alliance')], default='Character', max_length=12)),
                ('public_reason', models.CharField(blank=True, max_length=512, null=True)),
                ('reason', models.CharField(max_length=512)),
                ('issued_at', models.DateTimeField(auto_now_add=True)),
                ('revoked_at', models.DateTimeField(blank=True, default=None, null=True)),
                ('entity_alliance', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveonline.eveallianceinfo')),
                ('entity_character', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveonline.evecharacter')),
                ('entity_corporation', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveonline.evecorporationinfo')),
                ('issued_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='eveonline.evecharacter')),
                ('revoked_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='eveonline.evecharacter')),
            ],
            options={
                'verbose_name': 'Waitlist - Ban',
                'verbose_name_plural': 'Waitlist - Bans',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='CharacterBadges',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('granted_at', models.DateTimeField(auto_now_add=True)),
                ('badge', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.badge')),
                ('character', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveonline.evecharacter')),
                ('granted_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='eveonline.evecharacter')),
            ],
            options={
                'verbose_name': 'Waitlist - Character Badge',
                'verbose_name_plural': 'Waitlist - Character Badges',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='CharacterNote',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('note', models.TextField()),
                ('logged_at', models.DateTimeField(auto_now_add=True)),
                ('author', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='eveonline.evecharacter')),
                ('character', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveonline.evecharacter')),
            ],
            options={
                'verbose_name': 'Waitlist - Character Note',
                'verbose_name_plural': 'Waitlist - Character Note',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='CharacterRoles',
            fields=[
                ('character', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, serialize=False, to='eveonline.evecharacter')),
                ('granted_at', models.DateTimeField(auto_now_add=True)),
                ('granted_by', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='eveonline.evecharacter')),
            ],
            options={
                'verbose_name': 'Waitlist - Character Role',
                'verbose_name_plural': 'Waitlist - Character Roles',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Fitting',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('dna', models.TextField(max_length=1024, unique=True)),
                ('ship', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='corptools.eveitemtype', verbose_name='Ship')),
            ],
            options={
                'verbose_name': 'Waitlist - Waitlist Fit',
                'verbose_name_plural': 'Waitlist - Waitlist Fits',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='FittingHistory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('logged_at', models.DateTimeField(auto_now_add=True)),
                ('character', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveonline.evecharacter')),
                ('fit', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.fitting')),
            ],
            options={
                'verbose_name': 'Waitlist - Fitting History',
                'verbose_name_plural': 'Waitlist - Fitting Histories',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Fleet',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('is_updating', models.BooleanField(default=False)),
                ('boss', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveonline.evecharacter')),
            ],
            options={
                'verbose_name': 'Waitlist - Fleet',
                'verbose_name_plural': 'Waitlist - Fleets',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='FleetActivity',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('first_seen', models.DateTimeField(auto_now_add=True)),
                ('last_seen', models.DateTimeField(auto_now=True)),
                ('has_left', models.BooleanField(default=False)),
                ('is_boss', models.BooleanField(default=False)),
                ('character', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveonline.evecharacter')),
                ('fleet', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='incursions.fleet', verbose_name='')),
                ('ship', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='corptools.eveitemtype', verbose_name='Ship')),
            ],
            options={
                'verbose_name': 'Waitlist - Fleet Activity',
                'verbose_name_plural': 'Waitlist - Fleet Activities',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='FleetSquad',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('wing_id', models.BigIntegerField()),
                ('squad_id', models.BigIntegerField()),
            ],
            options={
                'verbose_name': 'Waitlist - Fleet Squad',
                'verbose_name_plural': 'Waitlist - Fleet Squad',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='ImplantSet',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('implants', models.CharField(max_length=255, unique=True)),
            ],
            options={
                'verbose_name': 'Waitlist - Waitlist Implant Set',
                'verbose_name_plural': 'Waitlist - Waitlist Implant Sets',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='Role',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=64, unique=True)),
                ('description', models.TextField(blank=True, null=True)),
                ('power', models.IntegerField(default=0, help_text='Bigger is better, user can assign a power level lower than this')),
            ],
            options={
                'verbose_name': 'Waitlist - Role',
                'verbose_name_plural': 'Waitlist - Roles',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='SkillCheck',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('min', models.IntegerField(default=3, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(5)])),
                ('elite', models.IntegerField(default=4, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(5)])),
                ('gold', models.IntegerField(default=5, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(5)])),
                ('skill', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='corptools.eveitemtype', verbose_name='Skill')),
            ],
            options={
                'verbose_name': 'Waitlist - Skill Check',
                'verbose_name_plural': 'Waitlist - Skill Check',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='SkillHistory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('old_level', models.IntegerField(validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(5)])),
                ('new_level', models.IntegerField(validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(5)])),
                ('logged_at', models.DateTimeField(auto_now_add=True)),
                ('character', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveonline.evecharacter')),
                ('skill', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='corptools.eveitemtype', verbose_name='Skill')),
            ],
            options={
                'verbose_name': 'Waitlist - Skill History',
                'verbose_name_plural': 'Waitlist - Skill Histories',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='WaitlistCategory',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
            ],
            options={
                'verbose_name': 'Waitlist - Category',
                'verbose_name_plural': 'Waitlist - Categories',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='WaitlistCategoryRule',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('ship', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='corptools.eveitemtype', verbose_name='Ship')),
                ('waitlist_category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.waitlistcategory', verbose_name='Waitlist Category')),
            ],
            options={
                'verbose_name': 'Waitlist - Category Rule',
                'verbose_name_plural': 'Waitlist - Category Rules',
                'default_permissions': (),
            },
        ),
        migrations.CreateModel(
            name='WaitlistEntryFit',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('approved', models.BooleanField(default=False)),
                ('tags', models.CharField(help_text='Comma Seperated List', max_length=255)),
                ('fit_analysis', models.TextField(blank=True, help_text='[] apparently', null=True)),
                ('review_comment', models.TextField(blank=True, null=True)),
                ('is_alt', models.BooleanField(default=False)),
                ('messagexup', models.TextField(blank=True, null=True)),
                ('cached_time_in_fleet', models.BigIntegerField()),
                ('category', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.waitlistcategory', verbose_name='Waitlist Category')),
                ('character', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveonline.evecharacter')),
                ('fit', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.fitting')),
                ('implant_set', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.implantset')),
            ],
            options={
                'verbose_name': 'Waitlist - Waitlist Entry Fit',
                'verbose_name_plural': 'Waitlist - Waitlist Entry Fit',
                'default_permissions': (),
            },
        ),
        migrations.AlterModelOptions(
            name='general',
            options={'default_permissions': (), 'managed': False, 'permissions': (('basic_waitlist', 'Can Access The Incursion Waitlist Tool'), ('waitlist_alts_view', 'Can View all of a users Alts ! Danger?'), ('waitlist_announcements_manage', 'can manage announcements'), ('waitlist_bages_view', 'Can View Badges'), ('waitlist_badges_manage', 'Can Manage Badges'), ('waitlist_bans_view', 'Can View Bans'), ('waitlist_bans_manage', 'Can Manage Bans'), ('waitlist_commanders_view', 'can view commanders'), ('waitlist_commanders_manage', 'can manage commanders'), ('waitlist_documentation_view', 'can view documentation'), ('waitlist_fleet_view', 'Can View Active Fleet composition'), ('waitlist_implants_view', 'Can View Implants'), ('waitlist_history_view', 'Can View History'), ('waitlist_notes_view', 'Can View Notes'), ('waitlist_notes_manage', 'Can Manage notes'), ('waitlist_search', 'Can Search'), ('waitlist_stats_view', 'Can View Fleet Stats'), ('waitlist_skills_view', 'Can view Skills'), ('waitlist_context_a', 'Add Context to Waitlist: Number of Pilots'), ('waitlist_context_b', 'Add Context to Waitlist: Ship Types'), ('waitlist_context_c', 'Add Context to Waitlist: Time in Waitlist'), ('waitlist_context_d', 'Add Context to Waitlist: Pilot Names'), ('waitlist_manage_waitlist', 'Manage Waitlist'), ('waitlist_manage_waitlist_approve_fits', 'Can Approve and Deny Fits on the WL'), ('waitlist_esi_show_info', 'Can open a Show Info Window ingame'), ('waitlist_esi_search', 'Will search ESI when using Search'))},
        ),
        migrations.AlterModelOptions(
            name='incursion',
            options={'default_permissions': (), 'verbose_name': 'Incursion', 'verbose_name_plural': 'Incursions'},
        ),
        migrations.AlterModelOptions(
            name='incursioninfluence',
            options={'default_permissions': (), 'verbose_name': 'IncursionInfluence', 'verbose_name_plural': 'IncursionInfluences'},
        ),
        migrations.AlterModelOptions(
            name='waitlist',
            options={'default_permissions': (), 'verbose_name': 'Waitlist - Waitlist', 'verbose_name_plural': 'Waitlist - Waitlist'},
        ),
        migrations.AlterModelOptions(
            name='waitlistentry',
            options={'default_permissions': (), 'verbose_name': 'Waitlist - Waitlist Entry', 'verbose_name_plural': 'Waitlist - Waitlist Entry'},
        ),
        migrations.AlterModelOptions(
            name='webhook',
            options={'default_permissions': (), 'verbose_name': 'Destination Webhook', 'verbose_name_plural': 'Destination Webhooks'},
        ),
        migrations.RemoveField(
            model_name='waitlist',
            name='close',
        ),
        migrations.RemoveField(
            model_name='waitlist',
            name='fc',
        ),
        migrations.RemoveField(
            model_name='waitlist',
            name='open',
        ),
        migrations.RemoveField(
            model_name='waitlistentry',
            name='note',
        ),
        migrations.RemoveField(
            model_name='waitlistentry',
            name='role',
        ),
        migrations.RemoveField(
            model_name='waitlistentry',
            name='ship',
        ),
        migrations.RemoveField(
            model_name='waitlistentry',
            name='time',
        ),
        migrations.RemoveField(
            model_name='waitlistentry',
            name='user',
        ),
        migrations.AddField(
            model_name='waitlist',
            name='is_open',
            field=models.BooleanField(default=False),
        ),
        migrations.AddField(
            model_name='waitlist',
            name='name',
            field=models.CharField(default='Waitlist', max_length=255),
        ),
        migrations.AddField(
            model_name='waitlistentry',
            name='joined_at',
            field=models.DateTimeField(auto_now_add=True, null=True),
        ),
        migrations.AddField(
            model_name='waitlistentry',
            name='main_character',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='eveonline.evecharacter'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='waitlistentry',
            name='waitlist',
            field=models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.waitlist'),
            preserve_default=False,
        ),
        migrations.DeleteModel(
            name='WaitlistShip',
        ),
        migrations.AddField(
            model_name='waitlistentryfit',
            name='waitlist_entry',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.waitlistentry'),
        ),
        migrations.AddField(
            model_name='fleetsquad',
            name='category',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.waitlistcategory', verbose_name='Waitlist Category'),
        ),
        migrations.AddField(
            model_name='fleetsquad',
            name='fleet',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.fleet'),
        ),
        migrations.AddField(
            model_name='fittinghistory',
            name='implant_set',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.implantset'),
        ),
        migrations.AddField(
            model_name='characterroles',
            name='role',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='incursion_role', to='incursions.role', verbose_name='Incursion Role'),
        ),
        migrations.AddField(
            model_name='approvedskills',
            name='skill_checks',
            field=models.ManyToManyField(related_name='+', to='incursions.skillcheck', verbose_name='Skill Checks'),
        ),
        migrations.AddField(
            model_name='approvedfitting',
            name='implants',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to='incursions.approvedimplantset'),
        ),
        migrations.AddField(
            model_name='approvedfitting',
            name='ship',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='corptools.eveitemtype', verbose_name='Ship'),
        ),
        migrations.AddField(
            model_name='activefleet',
            name='fleet',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='incursions.fleet'),
        ),
        migrations.AlterUniqueTogether(
            name='fleetsquad',
            unique_together={('fleet', 'category')},
        ),
    ]



from django.utils.translation import gettext_lazy as _

from allianceauth import hooks
from allianceauth.menu.hooks import MenuItemHook
from allianceauth.services.hooks import UrlHook

from incursions import urls


class IncursionsMenuItem(MenuItemHook):
    """This class ensures only authorized users will see the menu entry"""

    def __init__(self):
        # setup menu entry for sidebar
        MenuItemHook.__init__(
            self,
            _("Incursion Waitlist"),
            "fas fa-skull-crossbones fa-fw",
            "incursions:waitlist",
            navactive=["incursions:waitlist"],
        )

    def render(self, request):
        if request.user.has_perm("incursions.basic_incursions"):
            return MenuItemHook.render(self, request)
        return ""


@hooks.register('discord_cogs_hook')
def register_cogs() -> list[str]:
    return ["incursions.cogs.incursions"]


@hooks.register("menu_item_hook")
def register_menu() -> IncursionsMenuItem:
    return IncursionsMenuItem()


@hooks.register("url_hook")
def register_urls() -> UrlHook:
    return UrlHook(urls, "incursions", r"^incursions/")



from corptools.models import EveItemType
from ninja import NinjaAPI, Schema

from django.http import Http404

from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import CharacterSchema, HullSchema
from incursions.models.waitlist import (
    ActiveFleet, Fleet, FleetSquad, WaitlistCategory,
)
from incursions.providers import kick_all_fleet_members


class FleetStatusFleetSchema(Schema):
    id: int
    boss: CharacterSchema


class FleetStatusResponse(Schema):
    fleets : list[FleetStatusFleetSchema]


class FleetInfoSquadSchema(Schema):
    id: int
    name: str


class FleetInfoWingSchema(Schema):
    id: int
    name: str
    squads: list[FleetInfoSquadSchema]


class FleetInfoResponse(Schema):
    fleet_id: int
    wings: list[FleetInfoWingSchema]


class FleetMemberSchema(Schema):
    id: int
    name: str | None = None
    ship: HullSchema
    role: str


class FleetCompSquadMembersSchema(Schema):
    id: int
    name: str
    members: list[FleetMemberSchema]


class FleetCompWingSchema(Schema):
    id: int
    name: str
    squads: list[FleetCompSquadMembersSchema]
    member: FleetMemberSchema | None = None


class FleetCompResponse(Schema):
    wings: list[FleetCompWingSchema]
    id: int
    member: FleetMemberSchema | None = None


class FleetMembersMemberSchema(Schema):
    id: int
    name: str | None = None
    ship: HullSchema
    wl_category: str | None = None
    category: str | None = None
    role: str


class FleetMembersResponse(Schema):
    members: list[FleetMembersMemberSchema]


class RegisterRequest(Schema):
    character_id: int
    fleet_id: int
    assignments: dict[str, tuple[int, int]]


logger = get_extension_logger(__name__)

api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    FleetAPIEndpoints(api)


class FleetAPIEndpoints:

    tags = ["Fleets"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/status", response={200: FleetStatusResponse, 403: dict}, tags=self.tags)
        def fleet_status(request):
            if not request.user.has_perm("incursions.basic_waitlist"):
                return 403, {"error": "Permission denied"}

            fleets = ActiveFleet.get_solo().fleet

            return FleetStatusResponse(fleets=fleets)

        @api.get("/info", response={200: FleetInfoResponse, 403: dict}, tags=self.tags)
        def fleet_info(request, boss_character_id: int):
            if not request.user.has_perm("incursions.waitlist_manage_waitlist"):
                return 403, {"error": "Permission denied"}

            fleet = ActiveFleet.get_solo().fleet
            fleet_wings = fleet.get_fleet_wings()

            wings: list[FleetInfoWingSchema] = []
            for wing in fleet_wings:
                wing.append(FleetInfoWingSchema(
                    id=wing.id,
                    name=wing.name,
                    squads=[FleetInfoSquadSchema(id=squad.id, name=squad.name) for squad in wing.squads]
                ))

            return FleetInfoResponse(fleet_id=fleet.pk, wings=wings)

        @api.get("/fleetcomp", response=FleetCompResponse, tags=self.tags)
        def fleet_composition(request, boss_character_id: int):
            if not request.user.has_perm("incursions.waitlist_fleet_view"):
                return 403, {"error": "Permission denied"}

            fleet = ActiveFleet.get_solo().fleet
            fleet_members = fleet.get_fleet_members()
            fleet_wings = fleet.get_fleet_wings()

            wings: list[FleetCompWingSchema] = []
            for wing in fleet_wings:
                squads: list[FleetMemberSchema] = []
                for fleet_member in fleet_members:
                    if fleet_member.wing_id == wing.id:
                        squads.append(FleetMemberSchema(
                            id=fleet_member.squad_id,
                            name=EveCharacter.objects.get(id=fleet_member.id).character_name,
                            ship=HullSchema(id=fleet_member.ship_type_id, name=EveItemType.objects.get(id=fleet_member.ship_type_id)),
                            role=fleet_member.role_name
                        ))
                wings.append(FleetCompWingSchema(id=wing.id, name=wing.name, squads=squads))

            return FleetCompResponse(wings=wings, id=fleet.pk)

        @api.get("/fleet/members/{character_id}", response=FleetMembersResponse, tags=self.tags)
        def fleet_members(request, character_id: int):
            if not request.user.has_perm("incursions.waitlist_fleet_view"):
                return 403, {"error": "Permission denied"}

            try:
                fleet = Fleet.objects.filter(boss_id=character_id).get()
            except Fleet.DoesNotExist:
                return Http404("Fleet not found")

            fleet_members = fleet.get_fleet_members()
            wings = fleet.get_fleet_wings()

            wing_array = {}
            for wing in wings:  # We have to build a lil cute lookup table here off the ESI data, since its not indexed on ID
                wing_array.update((wing.id, wing.name))

            members: list[FleetMembersMemberSchema] = []

            for fleet_member in fleet_members:
                member = FleetMembersMemberSchema(
                    id=fleet_member.character_id,
                    name=fleet_member.character_name,
                    ship=HullSchema(id=fleet_member.ship_type_id, name=EveItemType.objects.get(id=fleet_member.ship_type_id).name),
                    wl_category=FleetSquad.objects.get(squad_id=fleet_member.squad_id).category,
                    category=f"{wing_array[fleet_member.wing_id]} - {fleet_member.squad_name}",  # from ESI wing/squad name
                    role=fleet_member.role,
                )
                members.append(member)

            return FleetMembersResponse(members=members)

        @api.post("/register", tags=self.tags)
        def register_fleet(request, body: RegisterRequest):
            if not request.user.has_perm("incursions.waitlist_manage_waitlist"):
                return 403, {"error": "Permission denied"}

            fleet, created = Fleet.objects.get_or_create(
                pk=body.fleet_id,
                defaults={"boss": EveCharacter.objects.get(id=body.character_id)},
            )

            active_fleet = ActiveFleet.get_solo()
            active_fleet.fleet = fleet
            fleet.save()

            for category, (wing_id, squad_id) in body.assignments.items():
                FleetSquad.objects.update_or_create(
                    fleet=fleet,
                    category=WaitlistCategory.objects.get(name=category),
                    defaults={
                        "wing_id": wing_id,
                        "squad_id": squad_id,
                    },
                )

            return "OK"

        @api.post("/close", tags=self.tags)
        def close_fleet(request, boss_character_id: int):
            if not request.user.has_perm("incursions.waitlist_manage_waitlist"):
                return 403, {"error": "Permission denied"}

            fleet = ActiveFleet.get_solo().fleet
            count = kick_all_fleet_members(boss_character_id=fleet.boss.character_id, fleet_id=fleet.pk)
            return f"Kicked {count} Fleet Members"



from enum import Enum

from ninja import Field, NinjaAPI, Schema

from allianceauth.services.hooks import get_extension_logger

from incursions.api.schema import HullSchema


class SkillLevelEnum(str, Enum):
    I = "I"  # noqa: E741
    II = "II"
    III = "III"
    IV = "IV"
    V = "V"


class BaseSkillPlanLevel(Schema):
    type: str


class FitLevel(Schema):
    type: str = "fit"
    hull: str
    fit: str


class SkillsLevel(Schema):
    type: str = "skills"
    from_: str = Field(..., alias="from")
    tier: str


class SkillLevelEntry(Schema):
    type: str = "skill"
    from_: str = Field(..., alias="from")
    level: SkillLevelEnum


class TankLevel(Schema):
    type: str = "tank"
    from_: str = Field(..., alias="from")


class SkillPlanSchema(Schema):
    name: str
    description: str
    alpha: bool = False
    plan: list[FitLevel | SkillsLevel | SkillLevelEntry | TankLevel] = Field(..., discriminator="type")


class SkillPlansResponsePlanSchema(Schema):
    source: str
    levels: list[tuple[int, int]]
    ships: list[HullSchema]


class SkillPlansResponse(Schema):
    plans: list[SkillPlansResponsePlanSchema]


logger = get_extension_logger(__name__)

api = NinjaAPI()


def setup(api: NinjaAPI) -> None:
    SkillPlansAPIEndpoints(api)


class SkillPlansAPIEndpoints:

    tags = ["Skill Plans"]

    def __init__(self, api: NinjaAPI) -> None:

        @api.get("/skills/plans", response={200: SkillPlansResponse, 403: dict}, tags=self.tags)
        def get_skill_plans(request):
            if not request.user.has_perm("incursions.basic_waitlist"):
                return 403, {"error": "Permission denied"}

            plans: list[SkillPlansResponsePlanSchema] = []
            plans.append(SkillPlansResponsePlanSchema(
                source=None,
                level=None,
                ships=None,
            ))

            return SkillPlansResponse(plans=plans)
